"""Configuration functionality for vcspull."""

from __future__ import annotations

import contextlib
import copy
import enum
import fnmatch
import logging
import os
import pathlib
import tempfile
import typing as t
from collections.abc import Callable

from libvcs.sync.git import GitRemote

from vcspull.validator import is_valid_config

from . import exc
from ._internal.config_reader import (
    ConfigReader,
    DuplicateAwareConfigReader,
    config_format_from_path,
)
from .types import ConfigDict, RawConfigDict, WorktreeConfigDict
from .util import get_config_dir, update_dict

log = logging.getLogger(__name__)


def expand_dir(
    dir_: pathlib.Path,
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
) -> pathlib.Path:
    """Return path with environmental variables and tilde ~ expanded.

    Parameters
    ----------
    dir_ : pathlib.Path
        Directory path to expand
    cwd : pathlib.Path, optional
        Current working dir (used to resolve relative paths). Defaults to
        :py:meth:`pathlib.Path.cwd`.

    Returns
    -------
    pathlib.Path
        Absolute directory path
    """
    dir_ = pathlib.Path(os.path.expandvars(str(dir_))).expanduser()
    if callable(cwd):
        cwd = cwd()

    if not dir_.is_absolute():
        dir_ = pathlib.Path(os.path.normpath(cwd / dir_))
        assert dir_ == pathlib.Path(cwd, dir_).resolve(strict=False)
    return dir_


def normalize_config_file_path(
    path: pathlib.Path,
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
) -> pathlib.Path:
    """Return absolute config file path without resolving symlinks.

    Symlink entry names are preserved intact so that downstream operations
    (e.g. atomic writes) can resolve them as needed, while the logical path
    is used for display and identity.

    Parameters
    ----------
    path : pathlib.Path
        Config file path to normalize.
    cwd : pathlib.Path, optional
        Current working dir (used to resolve relative paths). Defaults to
        :py:meth:`pathlib.Path.cwd`.

    Returns
    -------
    pathlib.Path
        Absolute config file path with symlink names preserved.

    Examples
    --------
    >>> normalize_config_file_path(pathlib.Path("~/cfg.yaml")).name
    'cfg.yaml'
    >>> normalize_config_file_path(
    ...     pathlib.Path("configs/vcspull.yaml"),
    ...     cwd=pathlib.Path("/tmp/project"),
    ... )  # doctest: +ELLIPSIS
    PosixPath('.../configs/vcspull.yaml')
    """
    path = pathlib.Path(os.path.expandvars(str(path))).expanduser()
    if callable(cwd):
        cwd = cwd()

    if not path.is_absolute():
        path = pathlib.Path(os.path.normpath(cwd / path))
    return path


def _validate_worktrees_config(
    worktrees_raw: t.Any,
    repo_name: str,
) -> list[WorktreeConfigDict]:
    """Validate and normalize worktrees configuration.

    Parameters
    ----------
    worktrees_raw : Any
        Raw worktrees configuration from YAML/JSON.
    repo_name : str
        Name of the parent repository (for error messages).

    Returns
    -------
    list[WorktreeConfigDict]
        Validated list of worktree configurations.

    Raises
    ------
    VCSPullException
        If the worktrees configuration is invalid.

    Examples
    --------
    Valid configuration with a tag:

    >>> from vcspull.config import _validate_worktrees_config
    >>> config = [{"dir": "../v1", "tag": "v1.0.0"}]
    >>> result = _validate_worktrees_config(config, "myrepo")
    >>> len(result)
    1
    >>> result[0]["dir"]
    '../v1'
    >>> result[0]["tag"]
    'v1.0.0'

    Valid configuration with a branch:

    >>> config = [{"dir": "../dev", "branch": "develop"}]
    >>> result = _validate_worktrees_config(config, "myrepo")
    >>> result[0]["branch"]
    'develop'

    Valid configuration with a commit:

    >>> config = [{"dir": "../fix", "commit": "abc123"}]
    >>> result = _validate_worktrees_config(config, "myrepo")
    >>> result[0]["commit"]
    'abc123'

    Error: worktrees must be a list:

    >>> _validate_worktrees_config("not-a-list", "myrepo")
    Traceback (most recent call last):
        ...
    vcspull.exc.VCSPullException: ...worktrees must be a list, got str

    Error: worktree entry must be a dict:

    >>> _validate_worktrees_config(["not-a-dict"], "myrepo")
    Traceback (most recent call last):
        ...
    vcspull.exc.VCSPullException: ...must be a dict, got str

    Error: missing required 'dir' field:

    >>> _validate_worktrees_config([{"tag": "v1.0.0"}], "myrepo")
    Traceback (most recent call last):
        ...
    vcspull.exc.VCSPullException: ...missing required 'dir' field

    Error: no ref type specified:

    >>> _validate_worktrees_config([{"dir": "../wt"}], "myrepo")
    Traceback (most recent call last):
        ...
    vcspull.exc.VCSPullException: ...must specify one of: tag, branch, or commit

    Error: empty ref value:

    >>> _validate_worktrees_config([{"dir": "../wt", "tag": ""}], "myrepo")
    Traceback (most recent call last):
        ...
    vcspull.exc.VCSPullException: ...empty ref value...

    Error: multiple refs specified:

    >>> _validate_worktrees_config(
    ...     [{"dir": "../wt", "tag": "v1", "branch": "main"}], "myrepo"
    ... )
    Traceback (most recent call last):
        ...
    vcspull.exc.VCSPullException: ...cannot specify multiple refs...
    """
    if not isinstance(worktrees_raw, list):
        msg = (
            f"Repository '{repo_name}': worktrees must be a list, "
            f"got {type(worktrees_raw).__name__}"
        )
        raise exc.VCSPullException(msg)

    validated: list[WorktreeConfigDict] = []

    for idx, wt in enumerate(worktrees_raw):
        if not isinstance(wt, dict):
            msg = (
                f"Repository '{repo_name}': worktree entry {idx} must be a dict, "
                f"got {type(wt).__name__}"
            )
            raise exc.VCSPullException(msg)

        # Validate required 'dir' field
        if "dir" not in wt or not wt["dir"]:
            msg = (
                f"Repository '{repo_name}': worktree entry {idx} "
                "missing required 'dir' field"
            )
            raise exc.VCSPullException(msg)

        if not isinstance(wt["dir"], str):
            msg = (
                f"Repository '{repo_name}': worktree entry {idx} "
                f"'dir' must be a string, got {type(wt['dir']).__name__}"
            )
            raise exc.VCSPullException(msg)

        # Validate exactly one ref type
        tag = wt.get("tag")
        branch = wt.get("branch")
        commit = wt.get("commit")

        refs_specified = sum(
            1 for ref in [tag, branch, commit] if ref is not None and ref != ""
        )
        empty_refs = sum(1 for ref in [tag, branch, commit] if ref == "")

        if refs_specified == 0 and empty_refs == 0:
            msg = (
                f"Repository '{repo_name}': worktree entry {idx} "
                "must specify one of: tag, branch, or commit"
            )
            raise exc.VCSPullException(msg)
        if refs_specified == 0 and empty_refs > 0:
            msg = (
                f"Repository '{repo_name}': worktree entry {idx} "
                "has empty ref value (tag, branch, or commit)"
            )
            raise exc.VCSPullException(msg)
        if refs_specified > 1:
            msg = (
                f"Repository '{repo_name}': worktree entry {idx} "
                "cannot specify multiple refs (tag, branch, commit)"
            )
            raise exc.VCSPullException(msg)

        # Validate ref types are strings
        for ref_name, ref_val in [("tag", tag), ("branch", branch), ("commit", commit)]:
            if ref_val is not None and not isinstance(ref_val, str):
                msg = (
                    f"Repository '{repo_name}': worktree entry {idx} "
                    f"'{ref_name}' must be a string, got {type(ref_val).__name__}"
                )
                raise exc.VCSPullException(msg)

        # Build validated worktree config
        wt_config: WorktreeConfigDict = {"dir": wt["dir"]}

        if tag:
            wt_config["tag"] = tag
        if branch:
            wt_config["branch"] = branch
        if commit:
            wt_config["commit"] = commit

        # Optional fields
        if "detach" in wt:
            wt_config["detach"] = wt["detach"]
        if "lock" in wt:
            wt_config["lock"] = wt["lock"]
        if "lock_reason" in wt:
            wt_config["lock_reason"] = wt["lock_reason"]

        validated.append(wt_config)

    return validated


def extract_repos(
    config: RawConfigDict,
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
) -> list[ConfigDict]:
    """Return expanded configuration.

    end-user configuration permit inline configuration shortcuts, expand to
    identical format for parsing.

    Parameters
    ----------
    config : dict
        the repo config in :py:class:`dict` format.
    cwd : pathlib.Path
        current working dir (for deciphering relative paths)

    Returns
    -------
    list : List of normalized repository information
    """
    configs: list[ConfigDict] = []
    if callable(cwd):
        cwd = cwd()

    for directory, repos in config.items():
        assert isinstance(repos, dict)
        for repo, repo_data in repos.items():
            conf: dict[str, t.Any] = {}

            """
            repo_name: http://myrepo.com/repo.git

            to

            repo_name: { url: 'http://myrepo.com/repo.git' }

            also assures the repo is a :py:class:`dict`.
            """

            if isinstance(repo_data, str):
                conf["url"] = repo_data
            else:
                conf = update_dict(conf, repo_data)

            if "repo" in conf:
                if "url" not in conf:
                    conf["url"] = conf.pop("repo")
                else:
                    conf.pop("repo", None)

            if "name" not in conf:
                conf["name"] = repo

            if "workspace_root" not in conf:
                conf["workspace_root"] = directory

            if "path" not in conf:
                conf["path"] = expand_dir(
                    pathlib.Path(expand_dir(pathlib.Path(directory), cwd=cwd))
                    / conf["name"],
                    cwd,
                )

            if "remotes" in conf:
                assert isinstance(conf["remotes"], dict)
                for remote_name, url in conf["remotes"].items():
                    if isinstance(url, GitRemote):
                        continue
                    if isinstance(url, str):
                        conf["remotes"][remote_name] = GitRemote(
                            name=remote_name,
                            fetch_url=url,
                            push_url=url,
                        )
                    elif isinstance(url, dict):
                        assert "push_url" in url
                        assert "fetch_url" in url
                        conf["remotes"][remote_name] = GitRemote(
                            name=remote_name,
                            **url,
                        )

            # Process worktrees configuration
            if "worktrees" in conf:
                worktrees_raw = conf["worktrees"]
                if worktrees_raw is not None:
                    repo_name_for_error = conf.get("name") or repo
                    validated_worktrees = _validate_worktrees_config(
                        worktrees_raw,
                        repo_name=repo_name_for_error,
                    )
                    conf["worktrees"] = validated_worktrees

            def is_valid_config_dict(val: t.Any) -> t.TypeGuard[ConfigDict]:
                assert isinstance(val, dict)
                return True

            assert is_valid_config_dict(conf)

            configs.append(conf)

    return configs


def find_home_config_files(
    filetype: list[str] | None = None,
) -> list[pathlib.Path]:
    """Return configs of ``.vcspull.{yaml,json}`` in user's home directory.

    The returned path preserves the logical home entry name so callers
    keep the config type implied by ``.yaml`` or ``.json`` even when the
    file is a symlink.

    Parameters
    ----------
    filetype : list of str, optional
        File types to search for (default ``["json", "yaml"]``)

    Returns
    -------
    list of pathlib.Path
        Absolute paths to discovered config files

    Examples
    --------
    >>> find_home_config_files()
    []
    """
    if filetype is None:
        filetype = ["json", "yaml"]
    configs: list[pathlib.Path] = []

    check_yaml = "yaml" in filetype
    check_json = "json" in filetype

    yaml_config = normalize_config_file_path(pathlib.Path("~/.vcspull.yaml"))
    has_yaml_config = check_yaml and yaml_config.exists()
    json_config = normalize_config_file_path(pathlib.Path("~/.vcspull.json"))
    has_json_config = check_json and json_config.exists()

    if not has_yaml_config and not has_json_config:
        log.debug(
            "No config file found. Create a .vcspull.yaml or .vcspull.json"
            " in your $HOME directory. http://vcspull.git-pull.com for a"
            " quickstart.",
        )
    else:
        if sum(filter(None, [has_json_config, has_yaml_config])) > 1:
            raise exc.MultipleConfigWarning
        if has_yaml_config:
            configs.append(yaml_config)
        if has_json_config:
            configs.append(json_config)

    return configs


def find_config_files(
    path: list[pathlib.Path] | pathlib.Path | None = None,
    match: list[str] | str | None = None,
    filetype: t.Literal["json", "yaml", "*"]
    | list[t.Literal["json", "yaml", "*"]]
    | None = None,
    include_home: bool = False,
) -> list[pathlib.Path]:
    """Return repos from a directory and match. Not recursive.

    Parameters
    ----------
    path : list
        list of paths to search
    match : list
        list of globs to search against
    filetype: list
        of filetypes to search against
    include_home : bool
        Include home configuration files

    Raises
    ------
    LoadConfigRepoConflict :
        There are two configs that have same path and name with different repo urls.

    Returns
    -------
    list :
        list of absolute paths to config files.
    """
    if filetype is None:
        filetype = ["json", "yaml"]
    if match is None:
        match = ["*"]
    config_files = []
    if path is None:
        path = get_config_dir()

    if include_home is True:
        config_files.extend(find_home_config_files())

    if isinstance(path, list):
        for p in path:
            config_files.extend(find_config_files(p, match, filetype))
            return config_files
    else:
        path = path.expanduser()
        if isinstance(match, list):
            for m in match:
                config_files.extend(find_config_files(path, m, filetype))
        elif isinstance(filetype, list):
            for f in filetype:
                config_files.extend(find_config_files(path, match, f))
        else:
            match = f"{match}.{filetype}"
            config_files = list(path.glob(match))

    return config_files


def load_configs(
    files: list[pathlib.Path],
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
    *,
    merge_duplicates: bool = True,
) -> list[ConfigDict]:
    """Return repos from a list of files.

    Parameters
    ----------
    files : list
        paths to config file
    cwd : pathlib.Path
        current path (pass down for :func:`extract_repos`

    Returns
    -------
    list of dict :
        expanded config dict item

    Todo
    ----
    Validate scheme, check for duplicate destinations, VCS urls
    """
    repos: list[ConfigDict] = []
    if callable(cwd):
        cwd = cwd()

    for file in files:
        if isinstance(file, str):
            file = pathlib.Path(file)
        assert isinstance(file, pathlib.Path)

        config_content, duplicate_roots, _top_level_items = (
            DuplicateAwareConfigReader.load_with_duplicates(file)
        )

        if merge_duplicates:
            (
                config_content,
                merge_conflicts,
                _merge_change_count,
                merge_details,
            ) = merge_duplicate_workspace_roots(config_content, duplicate_roots)

            for conflict in merge_conflicts:
                log.warning("%s: %s", file, conflict)

            for root_label, occurrence_count in merge_details:
                duplicate_count = max(occurrence_count - 1, 0)
                if duplicate_count == 0:
                    continue
                plural = "entry" if duplicate_count == 1 else "entries"
                log.info(
                    "%s: merged %d duplicate %s for workspace root '%s'",
                    file,
                    duplicate_count,
                    plural,
                    root_label,
                )
        elif duplicate_roots:
            duplicate_list = ", ".join(sorted(duplicate_roots.keys()))
            log.warning(
                "%s: duplicate workspace roots detected (%s); keeping last occurrences",
                file,
                duplicate_list,
            )

        assert is_valid_config(config_content)
        newrepos = extract_repos(config_content, cwd=cwd)

        if not repos:
            repos.extend(newrepos)
            continue

        dupes = detect_duplicate_repos(repos, newrepos)

        if len(dupes) > 0:
            msg = ("repos with same path + different VCS detected!", dupes)
            raise exc.VCSPullException(msg)
        repos.extend(newrepos)

    return repos


def detect_duplicate_repos(
    config1: list[ConfigDict],
    config2: list[ConfigDict],
) -> list[tuple[ConfigDict, ConfigDict]]:
    """Return duplicate repos dict if repo_dir same and vcs different.

    Parameters
    ----------
    config1 : list[ConfigDict]

    config2 : list[ConfigDict]

    Returns
    -------
    list[tuple[ConfigDict, ConfigDict]]
        List of duplicate tuples
    """
    if not config1:
        return []

    dupes: list[tuple[ConfigDict, ConfigDict]] = []

    repo_dirs = {
        pathlib.Path(repo["path"]).parent / repo["name"]: repo for repo in config1
    }
    repo_dirs_2 = {
        pathlib.Path(repo["path"]).parent / repo["name"]: repo for repo in config2
    }

    for repo_dir, repo in repo_dirs.items():
        if repo_dir in repo_dirs_2:
            dupes.append((repo, repo_dirs_2[repo_dir]))

    return dupes


def in_dir(
    config_dir: pathlib.Path | None = None,
    extensions: list[str] | None = None,
) -> list[str]:
    """Return a list of configs in ``config_dir``.

    Parameters
    ----------
    config_dir : str
        directory to search
    extensions : list
        filetypes to check (e.g. ``['.yaml', '.json']``).

    Returns
    -------
    list
    """
    if extensions is None:
        extensions = [".yml", ".yaml", ".json"]
    if config_dir is None:
        config_dir = get_config_dir()

    return [
        path.name
        for path in config_dir.iterdir()
        if is_config_file(path.name, extensions) and not path.name.startswith(".")
    ]


def filter_repos(
    config: list[ConfigDict],
    path: pathlib.Path | t.Literal["*"] | str | None = None,
    vcs_url: str | None = None,
    name: str | None = None,
) -> list[ConfigDict]:
    """Return a :py:obj:`list` list of repos from (expanded) config file.

    path, vcs_url and name all support fnmatch.

    Parameters
    ----------
    config : dict
        the expanded repo config in :py:class:`dict` format.
    path : str, Optional
        directory of checkout location, fnmatch pattern supported
    vcs_url : str, Optional
        url of vcs remote, fn match pattern supported
    name : str, Optional
        project name, fnmatch pattern supported

    Returns
    -------
    list :
        Repos
    """
    repo_list: list[ConfigDict] = []

    if path:
        repo_list.extend(
            [
                r
                for r in config
                if fnmatch.fnmatch(str(pathlib.Path(r["path"]).parent), str(path))
            ],
        )

    if vcs_url:
        repo_list.extend(
            r
            for r in config
            if fnmatch.fnmatch(str(r.get("url", r.get("repo"))), vcs_url)
        )

    if name:
        repo_list.extend(
            [r for r in config if fnmatch.fnmatch(str(r.get("name")), name)],
        )

    return repo_list


def is_config_file(
    filename: str,
    extensions: list[str] | str | None = None,
) -> bool:
    """Return True if file has a valid config file type.

    Parameters
    ----------
    filename : str
        filename to check (e.g. ``mysession.json``).
    extensions : list or str
        filetypes to check (e.g. ``['.yaml', '.json']``).

    Returns
    -------
    bool : True if is a valid config file type
    """
    if extensions is None:
        extensions = [".yml", ".yaml", ".json"]
    extensions = [extensions] if isinstance(extensions, str) else extensions
    return any(filename.endswith(e) for e in extensions)


def _atomic_write(target: pathlib.Path, content: str) -> None:
    """Write content to a file atomically via temp-file-then-rename.

    If *target* is a symbolic link the write goes through the symlink:
    the temporary file is created next to the resolved destination and
    the rename replaces the resolved path, leaving the symlink intact.

    Parameters
    ----------
    target : pathlib.Path
        Destination file path (may be a symlink)
    content : str
        Content to write

    Examples
    --------
    >>> import pathlib, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     p = pathlib.Path(tmp) / "test.txt"
    ...     _atomic_write(p, "hello")
    ...     p.read_text(encoding="utf-8")
    'hello'

    Symlinks are preserved — the real target is updated:

    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     real = pathlib.Path(tmp) / "real.txt"
    ...     _ = real.write_text("old", encoding="utf-8")
    ...     link = pathlib.Path(tmp) / "link.txt"
    ...     link.symlink_to(real)
    ...     _atomic_write(link, "new")
    ...     link.is_symlink(), link.read_text(encoding="utf-8")
    (True, 'new')
    """
    # Resolve symlinks so the temp file lives next to the real
    # destination and the rename replaces the real file, not the symlink.
    resolved = target.resolve()

    original_mode: int | None = None
    if resolved.exists():
        original_mode = resolved.stat().st_mode

    fd, tmp_path = tempfile.mkstemp(
        dir=resolved.parent,
        prefix=f".{resolved.name}.",
        suffix=".tmp",
    )
    try:
        with os.fdopen(fd, "w", encoding="utf-8") as f:
            f.write(content)
        if original_mode is not None:
            pathlib.Path(tmp_path).chmod(original_mode)
        pathlib.Path(tmp_path).replace(resolved)
    except BaseException:
        # Clean up the temp file on any failure
        with contextlib.suppress(OSError):
            pathlib.Path(tmp_path).unlink()
        raise


def save_config_yaml(config_file_path: pathlib.Path, data: dict[t.Any, t.Any]) -> None:
    """Save configuration data to a YAML file.

    Parameters
    ----------
    config_file_path : pathlib.Path
        Path to the configuration file to write
    data : dict
        Configuration data to save

    Examples
    --------
    >>> import pathlib, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     p = pathlib.Path(tmp) / "cfg.yaml"
    ...     save_config_yaml(p, {"~/code/": {"myrepo": "git+https://example.com/repo.git"}})
    ...     "myrepo" in p.read_text(encoding="utf-8")
    True
    """
    yaml_content = ConfigReader._dump(
        fmt="yaml",
        content=data,
        indent=2,
    )
    _atomic_write(config_file_path, yaml_content)


def save_config_json(config_file_path: pathlib.Path, data: dict[t.Any, t.Any]) -> None:
    """Save configuration data to a JSON file.

    Parameters
    ----------
    config_file_path : pathlib.Path
        Path to the configuration file to write
    data : dict
        Configuration data to save

    Examples
    --------
    >>> import json, pathlib, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     p = pathlib.Path(tmp) / "cfg.json"
    ...     save_config_json(p, {"~/code/": {"myrepo": "git+https://example.com/repo.git"}})
    ...     loaded = json.loads(p.read_text(encoding="utf-8"))
    ...     "~/code/" in loaded
    True
    """
    json_content = ConfigReader._dump(
        fmt="json",
        content=data,
        indent=2,
    )
    _atomic_write(config_file_path, json_content)


def save_config(config_file_path: pathlib.Path, data: dict[t.Any, t.Any]) -> None:
    """Save configuration data, dispatching by file extension.

    Parameters
    ----------
    config_file_path : pathlib.Path
        Path to the configuration file to write
    data : dict
        Configuration data to save

    Examples
    --------
    >>> import pathlib, tempfile, json
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     p = pathlib.Path(tmp) / "test.json"
    ...     save_config(p, {"~/code/": {"repo": {"repo": "git+https://x"}}})
    ...     loaded = json.loads(p.read_text(encoding="utf-8"))
    ...     loaded["~/code/"]["repo"]["repo"]
    'git+https://x'

    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     p = pathlib.Path(tmp) / "test.yaml"
    ...     save_config(p, {"~/code/": {"repo": {"repo": "git+https://x"}}})
    ...     "repo" in p.read_text(encoding="utf-8")
    True
    """
    if config_format_from_path(config_file_path) == "json":
        save_config_json(config_file_path, data)
    else:
        save_config_yaml(config_file_path, data)


def save_config_yaml_with_items(
    config_file_path: pathlib.Path,
    items: list[tuple[str, t.Any]],
) -> None:
    """Persist configuration data while preserving duplicate top-level sections.

    Unlike :func:`save_config_yaml`, which loses duplicate keys when given a
    plain ``dict``, this function accepts ordered ``(label, data)`` pairs so
    that two workspace-root entries with the same key are each serialised as a
    separate YAML block.

    Parameters
    ----------
    config_file_path : pathlib.Path
        Destination config file (may be a symlink; the real target is updated).
    items : list of tuple[str, Any]
        Ordered ``(section_label, section_data)`` pairs. Each pair becomes one
        YAML document block in the output.

    Examples
    --------
    >>> import pathlib, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     p = pathlib.Path(tmp) / "cfg.yaml"
    ...     save_config_yaml_with_items(p, [
    ...         ("~/code/", {"flask": "git+https://github.com/pallets/flask.git"}),
    ...         ("~/code/", {"django": "git+https://github.com/django/django.git"}),
    ...     ])
    ...     content = p.read_text(encoding="utf-8")
    ...     "flask" in content and "django" in content
    True
    """
    documents: list[str] = []

    for label, section in items:
        dumped = ConfigReader._dump(
            fmt="yaml",
            content={label: section},
            indent=2,
        ).rstrip()
        if dumped:
            documents.append(dumped)

    yaml_content = "\n".join(documents)
    if yaml_content:
        yaml_content += "\n"

    _atomic_write(config_file_path, yaml_content)


_VALID_OPS: frozenset[str] = frozenset({"import", "add", "discover", "fmt", "merge"})
"""Valid operation names for pin checking."""


def is_pinned_for_op(entry: t.Any, op: str) -> bool:
    """Return ``True`` if the repo config entry is pinned for *op*.

    Parameters
    ----------
    entry : Any
        Raw repo config value (string, dict, or ``None``).
    op : str
        Operation name: ``"import"``, ``"add"``, ``"discover"``,
        ``"fmt"``, or ``"merge"``.

    Examples
    --------
    Global pin applies to all ops:

    >>> is_pinned_for_op({"repo": "git+x", "options": {"pin": True}}, "import")
    True
    >>> is_pinned_for_op({"repo": "git+x", "options": {"pin": True}}, "fmt")
    True

    Per-op pin is scoped:

    >>> entry = {"repo": "git+x", "options": {"pin": {"import": True}}}
    >>> is_pinned_for_op(entry, "import")
    True
    >>> is_pinned_for_op(entry, "fmt")
    False

    ``allow_overwrite: false`` is shorthand for ``pin: {import: true}``
    (guards against ``--sync``):

    >>> entry2 = {"repo": "git+x", "options": {"allow_overwrite": False}}
    >>> is_pinned_for_op(entry2, "import")
    True
    >>> is_pinned_for_op(entry2, "add")
    False

    Plain string entries and entries without options are never pinned:

    >>> is_pinned_for_op("git+x", "import")
    False
    >>> is_pinned_for_op({"repo": "git+x"}, "import")
    False

    Explicit false is not pinned:

    >>> is_pinned_for_op({"repo": "git+x", "options": {"pin": False}}, "import")
    False

    String values for pin (not bool) are ignored — not pinned:

    >>> is_pinned_for_op({"repo": "git+x", "options": {"pin": "true"}}, "import")
    False

    Invalid op raises ValueError:

    >>> is_pinned_for_op(  # doctest: +IGNORE_EXCEPTION_DETAIL
    ...     {"repo": "git+x"}, "bogus"
    ... )
    Traceback (most recent call last):
        ...
    ValueError: Unknown op: 'bogus'
    """
    if op not in _VALID_OPS:
        msg = f"Unknown op: {op!r}"
        raise ValueError(msg)
    if not isinstance(entry, dict):
        return False
    opts = entry.get("options")
    if not isinstance(opts, dict):
        return False
    pin = opts.get("pin")
    if pin is True:
        return True
    if isinstance(pin, dict) and pin.get(op, False) is True:
        return True
    return op == "import" and opts.get("allow_overwrite", True) is False


def get_pin_reason(entry: t.Any) -> str | None:
    """Return the human-readable pin reason from a repo config entry.

    Non-string values are coerced to :func:`str` so callers can safely
    interpolate the result into log messages.

    Examples
    --------
    >>> entry = {"repo": "git+x", "options": {"pin": True, "pin_reason": "pinned"}}
    >>> get_pin_reason(entry)
    'pinned'
    >>> get_pin_reason({"repo": "git+x"}) is None
    True
    >>> get_pin_reason("git+x") is None
    True

    Non-string pin_reason is coerced:

    >>> get_pin_reason({"repo": "git+x", "options": {"pin_reason": 42}})
    '42'
    """
    if not isinstance(entry, dict):
        return None
    opts = entry.get("options")
    if not isinstance(opts, dict):
        return None
    reason = opts.get("pin_reason")
    if reason is None:
        return None
    return str(reason)


class MergeAction(enum.Enum):
    """Action for resolving a duplicate workspace-root repo conflict."""

    KEEP_EXISTING = "keep_existing"
    """First occurrence wins (standard behavior)."""

    KEEP_INCOMING = "keep_incoming"
    """Incoming pinned entry displaces unpinned existing entry."""


def _classify_merge_action(
    existing_entry: t.Any,
    incoming_entry: t.Any,
) -> MergeAction:
    """Classify the merge conflict resolution action.

    Parameters
    ----------
    existing_entry : Any
        The entry already stored (first occurrence).
    incoming_entry : Any
        The duplicate entry being merged in.

    Examples
    --------
    Neither pinned — keep existing (first-occurrence-wins):

    >>> _classify_merge_action({"repo": "git+a"}, {"repo": "git+b"})
    <MergeAction.KEEP_EXISTING: 'keep_existing'>

    Incoming is pinned — incoming wins:

    >>> _classify_merge_action(
    ...     {"repo": "git+a"},
    ...     {"repo": "git+b", "options": {"pin": True}},
    ... )
    <MergeAction.KEEP_INCOMING: 'keep_incoming'>

    Existing is pinned — existing wins regardless:

    >>> _classify_merge_action(
    ...     {"repo": "git+a", "options": {"pin": True}},
    ...     {"repo": "git+b"},
    ... )
    <MergeAction.KEEP_EXISTING: 'keep_existing'>

    Both pinned — first-occurrence-wins:

    >>> _classify_merge_action(
    ...     {"repo": "git+a", "options": {"pin": True}},
    ...     {"repo": "git+b", "options": {"pin": True}},
    ... )
    <MergeAction.KEEP_EXISTING: 'keep_existing'>
    """
    existing_pinned = is_pinned_for_op(existing_entry, "merge")
    incoming_pinned = is_pinned_for_op(incoming_entry, "merge")
    if incoming_pinned and not existing_pinned:
        return MergeAction.KEEP_INCOMING
    return MergeAction.KEEP_EXISTING


def merge_duplicate_workspace_root_entries(
    label: str,
    occurrences: list[t.Any],
) -> tuple[t.Any, list[str], int]:
    """Merge duplicate entries for a single workspace root."""
    conflicts: list[str] = []
    change_count = max(len(occurrences) - 1, 0)

    if not occurrences:
        return {}, conflicts, change_count

    if not all(isinstance(entry, dict) for entry in occurrences):
        conflicts.append(
            (
                f"Workspace root '{label}' contains duplicate entries that are not "
                "mappings. Keeping the last occurrence."
            ),
        )
        return occurrences[-1], conflicts, change_count

    merged: dict[str, t.Any] = {}

    for entry in occurrences:
        assert isinstance(entry, dict)
        for repo_name, repo_config in entry.items():
            if repo_name not in merged:
                merged[repo_name] = copy.deepcopy(repo_config)
            elif merged[repo_name] != repo_config:
                action = _classify_merge_action(merged[repo_name], repo_config)
                if action == MergeAction.KEEP_INCOMING:
                    merged[repo_name] = copy.deepcopy(repo_config)
                    reason = get_pin_reason(repo_config)
                    suffix = f" ({reason})" if reason else ""
                    conflicts.append(
                        f"'{label}': pinned entry for '{repo_name}'"
                        f" displaced earlier definition{suffix}"
                    )
                else:  # KEEP_EXISTING
                    reason = get_pin_reason(merged[repo_name])
                    qualifier = (
                        "pinned "
                        if is_pinned_for_op(merged[repo_name], "merge")
                        else ""
                    )
                    suffix = f" ({reason})" if reason else ""
                    conflicts.append(
                        f"'{label}': keeping {qualifier}entry for '{repo_name}'{suffix}"
                    )

    return merged, conflicts, change_count


def merge_duplicate_workspace_roots(
    config_data: dict[str, t.Any],
    duplicate_roots: dict[str, list[t.Any]],
) -> tuple[dict[str, t.Any], list[str], int, list[tuple[str, int]]]:
    """Merge duplicate workspace root sections captured during load."""
    if not duplicate_roots:
        return copy.deepcopy(config_data), [], 0, []

    merged_config = copy.deepcopy(config_data)
    conflicts: list[str] = []
    change_count = 0
    details: list[tuple[str, int]] = []

    for label, occurrences in duplicate_roots.items():
        (
            merged_value,
            entry_conflicts,
            entry_changes,
        ) = merge_duplicate_workspace_root_entries(
            label,
            occurrences,
        )
        merged_config[label] = merged_value
        conflicts.extend(entry_conflicts)
        change_count += entry_changes
        details.append((label, len(occurrences)))

    return merged_config, conflicts, change_count, details


def canonicalize_workspace_path(
    label: str,
    *,
    cwd: pathlib.Path | None = None,
) -> pathlib.Path:
    """Convert a workspace root label to an absolute canonical path."""
    cwd = cwd or pathlib.Path.cwd()
    label_path = pathlib.Path(label)
    return expand_dir(label_path, cwd=cwd)


def workspace_root_label(
    workspace_path: pathlib.Path,
    *,
    cwd: pathlib.Path | None = None,
    home: pathlib.Path | None = None,
    preserve_cwd_label: bool = True,
) -> str:
    """Create a normalized label for a workspace root path."""
    cwd = cwd or pathlib.Path.cwd()
    home = home or pathlib.Path.home()

    if preserve_cwd_label and workspace_path == cwd:
        return "./"

    if workspace_path == home:
        return "~/"

    try:
        relative_to_home = workspace_path.relative_to(home)
        label = f"~/{relative_to_home.as_posix()}"
    except ValueError:
        label = workspace_path.as_posix()

    if label != "./" and not label.endswith("/"):
        label += "/"

    return label


def normalize_workspace_roots(
    config_data: dict[str, t.Any],
    *,
    cwd: pathlib.Path | None = None,
    home: pathlib.Path | None = None,
    preserve_cwd_label: bool = True,
) -> tuple[dict[str, t.Any], dict[pathlib.Path, str], list[str], int]:
    """Normalize workspace root labels and merge duplicate sections."""
    cwd = cwd or pathlib.Path.cwd()
    home = home or pathlib.Path.home()

    normalized: dict[str, t.Any] = {}
    path_to_label: dict[pathlib.Path, str] = {}
    conflicts: list[str] = []
    change_count = 0

    for label, value in config_data.items():
        canonical_path = canonicalize_workspace_path(label, cwd=cwd)
        normalized_label = path_to_label.get(canonical_path)

        if normalized_label is None:
            normalized_label = workspace_root_label(
                canonical_path,
                cwd=cwd,
                home=home,
                preserve_cwd_label=preserve_cwd_label,
            )
            path_to_label[canonical_path] = normalized_label

            if isinstance(value, dict):
                normalized[normalized_label] = copy.deepcopy(value)
            else:
                normalized[normalized_label] = value

            if normalized_label != label:
                change_count += 1
        else:
            change_count += 1
            existing_value = normalized.get(normalized_label)

            if isinstance(existing_value, dict) and isinstance(value, dict):
                for repo_name, repo_config in value.items():
                    if repo_name not in existing_value:
                        existing_value[repo_name] = copy.deepcopy(repo_config)
                        change_count += 1
                    elif existing_value[repo_name] != repo_config:
                        conflict_message = (
                            f"Workspace root '{label}' contains conflicting "
                            "definitions for repository '{repo}'. Keeping the existing "
                            "entry."
                        )
                        conflicts.append(
                            conflict_message.format(
                                label=normalized_label,
                                repo=repo_name,
                            ),
                        )
            elif existing_value != value:
                conflict_message = (
                    f"Workspace root '{label}' contains conflicting non-dictionary "
                    "values. Keeping the existing entry."
                )
                conflicts.append(conflict_message.format(label=normalized_label))

    return normalized, path_to_label, conflicts, change_count



"""GitHub repository importer for vcspull."""

from __future__ import annotations

import logging
import typing as t
import urllib.parse

from .base import (
    HTTPClient,
    ImportMode,
    ImportOptions,
    RemoteRepo,
    filter_repo,
    get_token_from_env,
)

log = logging.getLogger(__name__)

GITHUB_API_URL = "https://api.github.com"
DEFAULT_PER_PAGE = 100
# GitHub search API limits results to 1000; exceeding this causes HTTP 422.
SEARCH_MAX_RESULTS = 1000


class GitHubImporter:
    """Importer for GitHub repositories.

    Supports three modes:
    - USER: Fetch repositories for a user
    - ORG: Fetch repositories for an organization
    - SEARCH: Search for repositories by query

    Examples
    --------
    >>> importer = GitHubImporter()
    >>> importer.service_name
    'GitHub'
    """

    service_name: str = "GitHub"

    def __init__(
        self,
        token: str | None = None,
        base_url: str | None = None,
    ) -> None:
        """Initialize the GitHub importer.

        Parameters
        ----------
        token : str | None
            GitHub API token. If not provided, will try GITHUB_TOKEN env var.
        base_url : str | None
            Base URL for GitHub Enterprise. Defaults to api.github.com.

        Notes
        -----
        Authentication is optional for public repositories. For private
        repositories or higher rate limits, set ``GITHUB_TOKEN`` or ``GH_TOKEN``.

        Classic PAT: no scopes needed for public repos; ``repo`` scope for
        private. Fine-grained PAT: "Metadata: Read-only" for public; add
        "Contents: Read-only" for private repos.

        Create a token at https://github.com/settings/tokens.

        Examples
        --------
        >>> importer = GitHubImporter(token="fake")
        >>> importer.service_name
        'GitHub'
        """
        self._token = token or get_token_from_env("GITHUB_TOKEN", "GH_TOKEN")
        self._base_url = (base_url or GITHUB_API_URL).rstrip("/")

        # GitHub Enterprise needs /api/v3; public api.github.com does not
        api_url = self._base_url
        if base_url and "/api/" not in self._base_url:
            api_url = f"{self._base_url}/api/v3"

        self._client = HTTPClient(
            api_url,
            token=self._token,
            auth_header="Authorization",
            auth_prefix="Bearer",
            user_agent="vcspull",
        )

    @property
    def is_authenticated(self) -> bool:
        """Check if the importer has authentication configured.

        Returns
        -------
        bool
            True if a token is configured

        Examples
        --------
        >>> GitHubImporter(token="fake").is_authenticated
        True
        """
        return self._token is not None

    def fetch_repos(self, options: ImportOptions) -> t.Iterator[RemoteRepo]:
        """Fetch repositories from GitHub.

        Parameters
        ----------
        options : ImportOptions
            Import options

        Yields
        ------
        RemoteRepo
            Repository information

        Raises
        ------
        AuthenticationError
            When authentication fails
        RateLimitError
            When rate limit is exceeded
        NotFoundError
            When user/org is not found
        """
        if options.mode == ImportMode.USER:
            yield from self._fetch_user(options)
        elif options.mode == ImportMode.ORG:
            yield from self._fetch_org(options)
        elif options.mode == ImportMode.SEARCH:
            yield from self._fetch_search(options)

    def _fetch_user(self, options: ImportOptions) -> t.Iterator[RemoteRepo]:
        """Fetch repositories for a user.

        Parameters
        ----------
        options : ImportOptions
            Import options

        Yields
        ------
        RemoteRepo
            Repository information
        """
        target = urllib.parse.quote(options.target, safe="")
        endpoint = f"/users/{target}/repos"
        yield from self._paginate_repos(endpoint, options)

    def _fetch_org(self, options: ImportOptions) -> t.Iterator[RemoteRepo]:
        """Fetch repositories for an organization.

        Parameters
        ----------
        options : ImportOptions
            Import options

        Yields
        ------
        RemoteRepo
            Repository information
        """
        target = urllib.parse.quote(options.target, safe="")
        endpoint = f"/orgs/{target}/repos"
        yield from self._paginate_repos(endpoint, options)

    def _fetch_search(self, options: ImportOptions) -> t.Iterator[RemoteRepo]:
        """Search for repositories.

        Parameters
        ----------
        options : ImportOptions
            Import options

        Yields
        ------
        RemoteRepo
            Repository information
        """
        query_parts = [options.target]

        if options.language:
            query_parts.append(f"language:{options.language}")

        if options.min_stars > 0:
            query_parts.append(f"stars:>={options.min_stars}")

        query = " ".join(query_parts)
        endpoint = "/search/repositories"
        page = 1
        count = 0
        total_available: int | None = None

        while count < options.limit:
            # Always use DEFAULT_PER_PAGE to maintain consistent pagination offset.
            # Changing per_page between pages causes offset misalignment and duplicates.
            params: dict[str, str | int] = {
                "q": query,
                "per_page": DEFAULT_PER_PAGE,
                "page": page,
                "sort": "stars",
                "order": "desc",
            }

            data, headers = self._client.get(
                endpoint,
                params=params,
                service_name=self.service_name,
            )

            self._log_rate_limit(headers)

            total_count = data.get("total_count", 0)
            if page == 1:
                total_available = total_count
                if total_count > 1000:
                    log.warning(
                        "GitHub search returned %d total results but API limits "
                        "to 1000; consider narrowing your query",
                        total_count,
                    )

            items = data.get("items", [])
            if not items:
                break

            for item in items:
                if count >= options.limit:
                    break

                repo = self._parse_repo(item)
                if filter_repo(repo, options):
                    yield repo
                    count += 1

            # Check if there are more pages
            if len(items) < DEFAULT_PER_PAGE:
                break

            # GitHub search API caps at 1000 results
            if page * DEFAULT_PER_PAGE >= SEARCH_MAX_RESULTS:
                break

            page += 1

        # Warn if results were truncated by --limit
        if (
            count >= options.limit
            and total_available is not None
            and total_available > count
        ):
            log.warning(
                "Showing %d of %d repositories (use --limit 0 to fetch all)",
                count,
                total_available,
            )

    def _paginate_repos(
        self,
        endpoint: str,
        options: ImportOptions,
    ) -> t.Iterator[RemoteRepo]:
        """Paginate through repository listing endpoints.

        Parameters
        ----------
        endpoint : str
            API endpoint
        options : ImportOptions
            Import options

        Yields
        ------
        RemoteRepo
            Repository information
        """
        page = 1
        count = 0
        more_available = False

        while count < options.limit:
            # Always use DEFAULT_PER_PAGE to maintain consistent pagination offset.
            # Changing per_page between pages causes offset misalignment and duplicates.
            params: dict[str, str | int] = {
                "per_page": DEFAULT_PER_PAGE,
                "page": page,
                "sort": "updated",
                "direction": "desc",
            }

            data, headers = self._client.get(
                endpoint,
                params=params,
                service_name=self.service_name,
            )

            self._log_rate_limit(headers)

            if not data:
                break

            for idx, item in enumerate(data):
                if count >= options.limit:
                    # Remaining items on this page or a full page = more exist
                    more_available = (
                        idx < len(data) - 1 or len(data) == DEFAULT_PER_PAGE
                    )
                    break

                repo = self._parse_repo(item)
                if filter_repo(repo, options):
                    yield repo
                    count += 1

            # Boundary: limit reached on the last item of a full page
            if count >= options.limit and len(data) == DEFAULT_PER_PAGE:
                more_available = True
                break

            # Check if there are more pages
            if len(data) < DEFAULT_PER_PAGE:
                break

            page += 1

        # Warn if results were truncated by --limit
        # GitHub user/org endpoints don't return total count
        if count >= options.limit and more_available:
            log.warning(
                "Showing %d repositories; more may be available "
                "(use --limit 0 to fetch all)",
                count,
            )

    def _parse_repo(self, data: dict[str, t.Any]) -> RemoteRepo:
        """Parse GitHub API response into RemoteRepo.

        Parameters
        ----------
        data : dict
            GitHub API repository data

        Returns
        -------
        RemoteRepo
            Parsed repository information
        """
        return RemoteRepo(
            name=data.get("name", ""),
            clone_url=data.get("clone_url", ""),
            ssh_url=data.get("ssh_url", ""),
            html_url=data.get("html_url", ""),
            description=data.get("description"),
            language=data.get("language"),
            topics=tuple(data.get("topics") or []),
            stars=data.get("stargazers_count", 0),
            is_fork=data.get("fork", False),
            is_archived=data.get("archived", False),
            default_branch=data.get("default_branch", "main"),
            owner=(data.get("owner") or {}).get("login", ""),
        )

    def _log_rate_limit(self, headers: dict[str, str]) -> None:
        """Log rate limit information from response headers.

        Parameters
        ----------
        headers : dict[str, str]
            Response headers
        """
        remaining = headers.get("x-ratelimit-remaining")
        limit = headers.get("x-ratelimit-limit")

        if remaining is not None and limit is not None:
            try:
                remaining_int = int(remaining)
            except (ValueError, TypeError):
                return
            if remaining_int < 10:
                log.warning(
                    "GitHub API rate limit low: %s/%s remaining",
                    remaining,
                    limit,
                )
            else:
                log.debug(
                    "GitHub API rate limit: %s/%s remaining",
                    remaining,
                    limit,
                )



"""``vcspull import codecommit`` subcommand."""

from __future__ import annotations

import argparse
import logging

from vcspull._internal.remotes import CodeCommitImporter, DependencyError

from .._colors import Colors, get_color_mode
from .._formatter import VcspullHelpFormatter
from ._common import _create_shared_parent, _run_import

log = logging.getLogger(__name__)


def create_codecommit_subparser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
) -> None:
    """Register the ``codecommit`` (aliases ``cc``, ``aws``) service subcommand.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The subparsers action from the ``import`` parser.
    """
    parser = subparsers.add_parser(
        "codecommit",
        aliases=["cc", "aws"],
        help="import from AWS CodeCommit",
        parents=[_create_shared_parent()],
        formatter_class=VcspullHelpFormatter,
        description="Import repositories from AWS CodeCommit.",
    )
    parser.add_argument(
        "target",
        metavar="TARGET",
        nargs="?",
        default="",
        help="Optional substring filter for repository names",
    )
    parser.add_argument(
        "--region",
        dest="region",
        metavar="REGION",
        help="AWS region for CodeCommit",
    )
    parser.add_argument(
        "--profile",
        dest="profile",
        metavar="PROFILE",
        help="AWS profile for CodeCommit",
    )
    parser.set_defaults(import_handler=handle_codecommit)


def handle_codecommit(args: argparse.Namespace) -> int:
    """Handle ``vcspull import codecommit``.

    Parameters
    ----------
    args : argparse.Namespace
        Parsed CLI arguments.

    Returns
    -------
    int
        Exit code (0 = success).
    """
    if args.workspace is None:
        msg = "-w/--workspace is required"
        raise SystemExit(msg)

    colors = Colors(get_color_mode(getattr(args, "color", "auto")))

    try:
        importer = CodeCommitImporter(
            region=getattr(args, "region", None),
            profile=getattr(args, "profile", None),
        )
    except DependencyError as exc:
        log.error("%s %s", colors.error("\u2717"), exc)  # noqa: TRY400
        return 1

    return _run_import(
        importer,
        service_name="codecommit",
        target=getattr(args, "target", "") or "",
        workspace=args.workspace,
        mode="user",
        language=getattr(args, "language", None),
        topics=getattr(args, "topics", None),
        min_stars=getattr(args, "min_stars", 0),
        include_archived=getattr(args, "include_archived", False),
        include_forks=getattr(args, "include_forks", False),
        limit=getattr(args, "limit", 100),
        config_path_str=getattr(args, "config", None),
        dry_run=getattr(args, "dry_run", False),
        yes=getattr(args, "yes", False),
        output_json=getattr(args, "output_json", False),
        output_ndjson=getattr(args, "output_ndjson", False),
        color=getattr(args, "color", "auto"),
        use_https=getattr(args, "use_https", False),
        sync=getattr(args, "sync", False),
        prune=getattr(args, "prune", False),
        prune_untracked=getattr(args, "prune_untracked", False),
        import_source=f"codecommit:{getattr(args, 'target', '') or '*'}",
    )



"""Custom help formatter used by vcspull CLI."""

from __future__ import annotations

import argparse
import re
import typing as t

OPTIONS_EXPECTING_VALUE = {
    "-f",
    "--file",
    "-w",
    "--workspace",
    "--workspace-root",
    "--log-level",
    "--path",
    "--color",
    "--field",
    "--max-concurrent",
    "--name",
    "--url",
    "--region",
    "--profile",
    "--token",
    "-m",
    "--mode",
    "-l",
    "--language",
    "--topics",
    "--min-stars",
    "--limit",
}

OPTIONS_FLAG_ONLY = {
    "-h",
    "--help",
    "--write",
    "--all",
    "-a",
    "--recursive",
    "-r",
    "--yes",
    "-y",
    "--dry-run",
    "-n",
    "--json",
    "--ndjson",
    "--tree",
    "--detailed",
    "-d",
    "-i",
    "--ignore-case",
    "-S",
    "--smart-case",
    "-F",
    "--fixed-strings",
    "--word-regexp",
    "-v",
    "--invert-match",
    "--any",
    "--exit-on-error",
    "-x",
    "--fetch",
    "--long",
    "--offline",
    "--relative-paths",
    "--show-unchanged",
    "--summary-only",
    "--version",
    "-V",
    "--no-concurrent",
    "--sequential",
    "--no-merge",
    "--verbose",
    "--flatten-groups",
    "--archived",
    "--forks",
    "--https",
}


class _HelpTheme(t.Protocol):
    """Protocol describing the argparse color theme.

    Python 3.14+ sets ``self._theme`` on ``HelpFormatter`` instances via
    ``_colorize.get_theme().argparse``.  This protocol documents the
    attributes consumed by :meth:`VcspullHelpFormatter._fill_text` and
    :meth:`VcspullHelpFormatter._colorize_example_line`.
    """

    heading: str
    reset: str
    label: str
    long_option: str
    short_option: str
    prog: str
    action: str


class VcspullHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Extend argparse help colorization to example command sections.

    Python 3.14+ natively colorizes usage and option groups via
    ``_set_color()`` / ``_colorize.get_theme().argparse``.  This
    formatter hooks into the same ``_theme`` attribute to additionally
    colorize the "examples:" blocks in description text, applying
    distinct colors to program names, subcommands, long/short options,
    and option values.

    When ``_theme`` is ``None`` (older Python or ``NO_COLOR`` set),
    ``_fill_text`` falls through to the base class unchanged.
    """

    def _fill_text(self, text: str, width: int, indent: str) -> str:
        theme = t.cast("_HelpTheme | None", getattr(self, "_theme", None))
        if not text or theme is None:
            return super()._fill_text(text, width, indent)

        lines = text.splitlines(keepends=True)
        formatted_lines: list[str] = []
        in_examples_block = False
        expect_value = False

        for line in lines:
            if line.strip() == "":
                in_examples_block = False
                expect_value = False
                formatted_lines.append(f"{indent}{line}")
                continue

            has_newline = line.endswith("\n")
            stripped_line = line.rstrip("\n")
            leading_length = len(stripped_line) - len(stripped_line.lstrip(" "))
            leading = stripped_line[:leading_length]
            content = stripped_line[leading_length:]
            content_lower = content.lower()
            is_section_heading = (
                content_lower.endswith("examples:") and content_lower != "examples:"
            )

            if is_section_heading or content_lower == "examples:":
                formatted_content = f"{theme.heading}{content}{theme.reset}"
                in_examples_block = True
                expect_value = False
            elif in_examples_block:
                colored_content = self._colorize_example_line(
                    content,
                    theme=theme,
                    expect_value=expect_value,
                )
                expect_value = colored_content.expect_value
                formatted_content = colored_content.text
            else:
                formatted_content = stripped_line

            newline = "\n" if has_newline else ""
            formatted_lines.append(f"{indent}{leading}{formatted_content}{newline}")

        return "".join(formatted_lines)

    class _ColorizedLine(t.NamedTuple):
        text: str
        expect_value: bool

    def _colorize_example_line(
        self,
        content: str,
        *,
        theme: _HelpTheme,
        expect_value: bool,
    ) -> _ColorizedLine:
        parts: list[str] = []
        expecting_value = expect_value
        first_token = True
        colored_subcommand = False

        for match in re.finditer(r"\s+|\S+", content):
            token = match.group()
            if token.isspace():
                parts.append(token)
                continue

            if expecting_value:
                color = theme.label
                expecting_value = False
            elif token.startswith("--"):
                color = theme.long_option
                expecting_value = (
                    token not in OPTIONS_FLAG_ONLY and token in OPTIONS_EXPECTING_VALUE
                )
            elif token.startswith("-"):
                color = theme.short_option
                expecting_value = (
                    token not in OPTIONS_FLAG_ONLY and token in OPTIONS_EXPECTING_VALUE
                )
            elif first_token:
                color = theme.prog
            elif not colored_subcommand:
                color = theme.action
                colored_subcommand = True
            else:
                color = None

            first_token = False

            if color:
                parts.append(f"{color}{token}{theme.reset}")
            else:
                parts.append(token)

        return self._ColorizedLine(text="".join(parts), expect_value=expecting_value)



"""Typings for vcspull.

Configuration Object Graph
--------------------------

The user-facing ``.vcspull.yaml`` maps *workspace roots* (parent directories)
to named repositories.  For example::

    ~/study/c:
      cpython:
        repo: git+https://github.com/python/cpython.git
      tmux:
        repo: git+https://github.com/tmux/tmux.git

    ~/work/js:
      react:
        repo: https://github.com/facebook/react.git
      vite:
        repo: https://github.com/vitejs/vite.git

In Python we model this as:

``WorkspaceRoot`` - Mapping of repository name to its configuration
``WorkspaceRoots`` - Mapping of workspace root path to ``WorkspaceRoot``

When the configuration is parsed we preserve the original key string, but
``WorkspaceRoot`` terminology is used consistently across the codebase.
"""

from __future__ import annotations

import pathlib
import typing as t
from typing import TypeAlias

from typing_extensions import NotRequired, TypedDict

if t.TYPE_CHECKING:
    from libvcs._internal.types import StrPath, VCSLiteral
    from libvcs.sync.git import GitSyncRemoteDict


class WorktreeConfigDict(TypedDict):
    """Configuration for a single git worktree.

    Worktrees allow checking out multiple branches/tags/commits of a repository
    simultaneously in separate directories.

    Exactly one of ``tag``, ``branch``, or ``commit`` must be specified.

    Examples
    --------
    Tag worktree (immutable, detached HEAD)::

        {"dir": "../myproject-v1.0", "tag": "v1.0.0"}

    Branch worktree (updatable)::

        {"dir": "../myproject-dev", "branch": "develop"}

    Commit worktree (immutable, detached HEAD)::

        {"dir": "../myproject-abc", "commit": "abc123"}
    """

    dir: str
    """Path for the worktree (relative to workspace root or absolute)."""

    tag: NotRequired[str | None]
    """Tag to checkout (creates detached HEAD)."""

    branch: NotRequired[str | None]
    """Branch to checkout (can be updated/pulled)."""

    commit: NotRequired[str | None]
    """Commit SHA to checkout (creates detached HEAD)."""

    detach: NotRequired[bool | None]
    """Force detached HEAD. Default: True for tag/commit, False for branch."""

    lock: NotRequired[bool | None]
    """Lock the worktree to prevent accidental removal."""

    lock_reason: NotRequired[str | None]
    """Reason for locking. If provided, implies lock=True."""


RepoPinDict = TypedDict(
    "RepoPinDict",
    {
        "add": bool,
        "discover": bool,
        "fmt": bool,
        "import": bool,
        "merge": bool,
    },
    total=False,
)
"""Per-operation pin flags for a repository entry.

Unspecified keys default to ``False`` (not pinned).

Note: Distinct from ``WorktreeConfigDict.lock`` which prevents git worktree
removal at the filesystem level. ``RepoPinDict`` controls vcspull config
mutation policy only.

Examples
--------
Pin only import::

    options:
      pin:
        import: true

Pin import and fmt::

    options:
      pin:
        import: true
        fmt: true
"""


class RepoOptionsDict(TypedDict, total=False):
    """Mutation policy stored under the ``options:`` key in a repo entry.

    Note: ``pin`` here controls vcspull config mutation. It is distinct from
    ``WorktreeConfigDict.lock`` which prevents git worktree removal.

    Examples
    --------
    Pin all operations::

        options:
          pin: true
          pin_reason: "pinned to upstream"

    Pin only import (prevent ``--sync`` from replacing URL)::

        options:
          pin:
            import: true

    Shorthand form — equivalent to ``pin: {import: true}``::

        options:
          allow_overwrite: false
    """

    pin: bool | RepoPinDict
    """``True`` pins all ops; a mapping pins specific ops only.

    Unspecified keys in the mapping default to ``False`` (not pinned).
    """

    allow_overwrite: bool
    """If ``False``, shorthand for ``pin: {import: true}``.

    Pins only the import operation.
    """

    pin_reason: str | None
    """Human-readable reason shown in log output when an op is skipped due to pin."""


class RepoEntryDict(TypedDict):
    """Raw per-repository entry as written to .vcspull.yaml.

    Examples
    --------
    Minimal entry::

        repo: git+git@github.com:user/myrepo.git

    With pin options::

        repo: git+git@github.com:user/myrepo.git
        options:
          pin:
            import: true
          pin_reason: "pinned to company fork"
    """

    repo: str
    """VCS URL in vcspull format, e.g. ``git+git@github.com:user/repo.git``."""

    options: NotRequired[RepoOptionsDict]
    """Mutation policy. Nested under ``options:`` to avoid polluting VCS fields."""


class RawConfigDict(t.TypedDict):
    """Configuration dictionary without any type marshalling or variable resolution."""

    vcs: VCSLiteral
    name: str
    path: StrPath
    url: str
    remotes: GitSyncRemoteDict


RawConfigDir = dict[str, RawConfigDict]
RawConfig = dict[str, RawConfigDir]


class ConfigDict(TypedDict):
    """Configuration map for vcspull after shorthands and variables resolved."""

    vcs: VCSLiteral | None
    name: str
    path: pathlib.Path
    url: str
    workspace_root: str
    remotes: NotRequired[GitSyncRemoteDict | None]
    shell_command_after: NotRequired[list[str] | None]
    worktrees: NotRequired[list[WorktreeConfigDict] | None]
    options: NotRequired[RepoOptionsDict]


ConfigDir = dict[str, ConfigDict]
Config = dict[str, ConfigDir]

# ---------------------------------------------------------------------------
# Workspace root aliases
# ---------------------------------------------------------------------------

WorkspaceRoot = ConfigDir

WorkspaceRoots: TypeAlias = dict[pathlib.Path, WorkspaceRoot]



"""``vcspull import forgejo`` subcommand."""

from __future__ import annotations

import argparse

from vcspull._internal.remotes import GiteaImporter

from .._formatter import VcspullHelpFormatter
from ._common import (
    _create_mode_parent,
    _create_shared_parent,
    _create_target_parent,
    _create_token_parent,
    _run_import,
)


def create_forgejo_subparser(
    subparsers: argparse._SubParsersAction[argparse.ArgumentParser],
) -> None:
    """Register the ``forgejo`` service subcommand.

    Parameters
    ----------
    subparsers : argparse._SubParsersAction
        The subparsers action from the ``import`` parser.
    """
    parser = subparsers.add_parser(
        "forgejo",
        help="import from a Forgejo instance",
        parents=[
            _create_shared_parent(),
            _create_token_parent(),
            _create_mode_parent(),
            _create_target_parent(),
        ],
        formatter_class=VcspullHelpFormatter,
        description="Import repositories from a Forgejo instance.",
    )
    parser.add_argument(
        "--url",
        dest="base_url",
        metavar="URL",
        required=True,
        help="Base URL of the Forgejo instance (required)",
    )
    parser.set_defaults(import_handler=handle_forgejo)


def handle_forgejo(args: argparse.Namespace) -> int:
    """Handle ``vcspull import forgejo``.

    Parameters
    ----------
    args : argparse.Namespace
        Parsed CLI arguments.

    Returns
    -------
    int
        Exit code (0 = success).
    """
    if args.workspace is None:
        msg = "-w/--workspace is required"
        raise SystemExit(msg)

    importer = GiteaImporter(
        token=getattr(args, "token", None),
        base_url=args.base_url,
    )
    return _run_import(
        importer,
        service_name="forgejo",
        target=args.target,
        workspace=args.workspace,
        mode=args.mode,
        language=getattr(args, "language", None),
        topics=getattr(args, "topics", None),
        min_stars=getattr(args, "min_stars", 0),
        include_archived=getattr(args, "include_archived", False),
        include_forks=getattr(args, "include_forks", False),
        limit=getattr(args, "limit", 100),
        config_path_str=getattr(args, "config", None),
        dry_run=getattr(args, "dry_run", False),
        yes=getattr(args, "yes", False),
        output_json=getattr(args, "output_json", False),
        output_ndjson=getattr(args, "output_ndjson", False),
        color=getattr(args, "color", "auto"),
        use_https=getattr(args, "use_https", False),
        sync=getattr(args, "sync", False),
        prune=getattr(args, "prune", False),
        prune_untracked=getattr(args, "prune_untracked", False),
        import_source=f"forgejo:{args.target}",
    )



"""AWS CodeCommit repository importer for vcspull."""

from __future__ import annotations

import json
import logging
import subprocess
import typing as t

from .base import (
    AuthenticationError,
    ConfigurationError,
    DependencyError,
    ImportOptions,
    RemoteRepo,
    ServiceUnavailableError,
    filter_repo,
)

log = logging.getLogger(__name__)


class CodeCommitImporter:
    """Importer for AWS CodeCommit repositories.

    Uses AWS CLI to list and fetch repository information.
    Requires AWS CLI to be installed and configured.

    Examples
    --------
    >>> importer = CodeCommitImporter(region="us-east-1")
    >>> importer.service_name
    'CodeCommit'
    """

    service_name: str = "CodeCommit"

    def __init__(
        self,
        region: str | None = None,
        profile: str | None = None,
    ) -> None:
        """Initialize the CodeCommit importer.

        Parameters
        ----------
        region : str | None
            AWS region. If not provided, uses AWS CLI default.
        profile : str | None
            AWS profile name. If not provided, uses default profile.

        Notes
        -----
        Uses AWS CLI credentials (``aws configure``). No token environment
        variable is used. IAM policy must include
        ``codecommit:ListRepositories`` (resource ``*``) and
        ``codecommit:BatchGetRepositories``.

        Requires AWS CLI: ``pip install awscli``.
        """
        self._region = region
        self._profile = profile
        self._check_aws_cli()

    def _check_aws_cli(self) -> None:
        """Check if AWS CLI is installed and accessible.

        Raises
        ------
        DependencyError
            When AWS CLI is not installed
        """
        try:
            result = subprocess.run(
                ["aws", "--version"],
                capture_output=True,
                text=True,
                check=False,
            )
            if result.returncode != 0:
                msg = (
                    "AWS CLI not installed or not accessible. "
                    "Please install it with: pip install awscli"
                )
                raise DependencyError(msg, service=self.service_name)
        except FileNotFoundError as exc:
            msg = "AWS CLI not installed. Please install it with: pip install awscli"
            raise DependencyError(msg, service=self.service_name) from exc

    def _build_aws_command(self, *args: str) -> list[str]:
        """Build AWS CLI command with region and profile options.

        Parameters
        ----------
        *args : str
            AWS CLI arguments

        Returns
        -------
        list[str]
            Complete command list
        """
        cmd = ["aws", "--output", "json"]
        if self._region:
            cmd.extend(["--region", self._region])
        if self._profile:
            cmd.extend(["--profile", self._profile])
        cmd.extend(args)
        return cmd

    def _run_aws_command(self, *args: str) -> dict[str, t.Any]:
        """Run an AWS CLI command and return parsed JSON output.

        Parameters
        ----------
        *args : str
            AWS CLI arguments

        Returns
        -------
        dict
            Parsed JSON output

        Raises
        ------
        AuthenticationError
            When AWS credentials are missing or invalid
        ConfigurationError
            When region is invalid or endpoint unreachable
        """
        cmd = self._build_aws_command(*args)
        log.debug("Running: %s", " ".join(cmd))

        try:
            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
                timeout=60,
            )
        except FileNotFoundError as exc:
            msg = "AWS CLI not found"
            raise DependencyError(msg, service=self.service_name) from exc
        except subprocess.TimeoutExpired as exc:
            msg = "AWS CLI command timed out"
            raise ServiceUnavailableError(msg, service=self.service_name) from exc

        if result.returncode != 0:
            stderr = result.stderr.lower()
            if "unable to locate credentials" in stderr:
                msg = (
                    "AWS credentials not configured. Run 'aws configure' or "
                    "set AWS_ACCESS_KEY_ID/AWS_SECRET_ACCESS_KEY."
                )
                raise AuthenticationError(msg, service=self.service_name)
            if "could not connect to the endpoint" in stderr:
                msg = (
                    f"Could not connect to CodeCommit. Check your region setting. "
                    f"Error: {result.stderr}"
                )
                raise ConfigurationError(msg, service=self.service_name)
            if "invalid" in stderr and "region" in stderr:
                msg = f"Invalid AWS region. Error: {result.stderr}"
                raise ConfigurationError(msg, service=self.service_name)
            msg = f"AWS CLI error: {result.stderr}"
            raise ConfigurationError(msg, service=self.service_name)

        try:
            return json.loads(result.stdout) if result.stdout.strip() else {}
        except json.JSONDecodeError as exc:
            msg = f"Invalid JSON from AWS CLI: {result.stdout}"
            raise ConfigurationError(msg, service=self.service_name) from exc

    @property
    def is_authenticated(self) -> bool:
        """Check if AWS credentials are configured.

        Returns
        -------
        bool
            True if credentials appear to be configured
        """
        try:
            self._run_aws_command("sts", "get-caller-identity")
        except (AuthenticationError, ConfigurationError):
            return False
        else:
            return True

    def fetch_repos(self, options: ImportOptions) -> t.Iterator[RemoteRepo]:
        """Fetch repositories from AWS CodeCommit.

        Parameters
        ----------
        options : ImportOptions
            Import options (target is used as optional name filter)

        Yields
        ------
        RemoteRepo
            Repository information

        Raises
        ------
        AuthenticationError
            When AWS credentials are missing
        ConfigurationError
            When region is invalid
        DependencyError
            When AWS CLI is not installed
        """
        # List all repositories (paginate over nextToken)
        repositories: list[dict[str, t.Any]] = []
        next_token: str | None = None
        while True:
            cmd_args = ["codecommit", "list-repositories"]
            if next_token:
                cmd_args.extend(["--next-token", next_token])
            data = self._run_aws_command(*cmd_args)
            repositories.extend(data.get("repositories", []))
            next_token = data.get("nextToken")
            if not next_token:
                break

        if not repositories:
            return

        # Filter by name if target is provided
        if options.target:
            target_lower = options.target.lower()
            repositories = [
                r
                for r in repositories
                if target_lower in r.get("repositoryName", "").lower()
            ]

        # Batch get repository details (up to 25 at a time)
        count = 0
        batch_size = 25

        for i in range(0, len(repositories), batch_size):
            if count >= options.limit:
                break

            batch = repositories[i : i + batch_size]
            repo_names = [r["repositoryName"] for r in batch]

            # Get detailed info for batch
            details = self._run_aws_command(
                "codecommit",
                "batch-get-repositories",
                "--repository-names",
                *repo_names,
            )

            for repo_metadata in details.get("repositories", []):
                if count >= options.limit:
                    break

                repo = self._parse_repo(repo_metadata)
                if filter_repo(repo, options):
                    yield repo
                    count += 1

    def _parse_repo(self, data: dict[str, t.Any]) -> RemoteRepo:
        """Parse CodeCommit repository metadata into RemoteRepo.

        Parameters
        ----------
        data : dict
            CodeCommit repository metadata

        Returns
        -------
        RemoteRepo
            Parsed repository information
        """
        repo_name = data.get("repositoryName", "")
        account_id = data.get("accountId", "")

        # Build console URL
        region = self._region
        if not region:
            # Extract region from clone URL
            # (format: https://git-codecommit.{region}.amazonaws.com/...)
            clone_http = data.get("cloneUrlHttp", "")
            if "git-codecommit." in clone_http:
                region = clone_http.split("git-codecommit.")[1].split(".")[0]
            else:
                region = "us-east-1"
                log.debug(
                    "Could not determine region, defaulting to %s for console URL",
                    region,
                )
        html_url = (
            f"https://{region}.console.aws.amazon.com/codesuite/codecommit/"
            f"repositories/{repo_name}/browse"
        )

        return RemoteRepo(
            name=repo_name,
            clone_url=data.get("cloneUrlHttp", ""),
            ssh_url=data.get("cloneUrlSsh", ""),
            html_url=html_url,
            description=data.get("repositoryDescription"),
            language=None,  # CodeCommit doesn't track language
            topics=(),  # CodeCommit doesn't have topics
            stars=0,  # CodeCommit doesn't have stars
            is_fork=False,  # CodeCommit doesn't have forks
            is_archived=False,  # CodeCommit doesn't have archived state
            default_branch=data.get("defaultBranch", "main"),
            owner=account_id,
        )



"""Validation of vcspull configuration file."""

from __future__ import annotations

import pathlib
import typing as t

from vcspull.types import RawConfigDict


def is_valid_config(config: dict[str, t.Any]) -> t.TypeGuard[RawConfigDict]:
    """Return true and upcast if vcspull configuration file is valid."""
    if not isinstance(config, dict):
        return False

    for k, v in config.items():
        if k is None or v is None:
            return False

        if not isinstance(k, str) and not isinstance(k, pathlib.Path):
            return False

        if not isinstance(v, dict):
            return False

        for repo in v.values():
            if not isinstance(repo, (str, dict, pathlib.Path)):
                return False

            if isinstance(repo, dict) and "url" not in repo and "repo" not in repo:
                return False

    return True



"""Synchronization functionality for vcspull."""

from __future__ import annotations

import argparse
import asyncio
import contextlib
import logging
import os
import pathlib
import re
import subprocess
import sys
import typing as t
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from io import StringIO
from time import perf_counter

from libvcs._internal.shortcuts import create_project
from libvcs._internal.types import VCSLiteral
from libvcs.sync.git import GitSync
from libvcs.sync.hg import HgSync
from libvcs.sync.svn import SvnSync
from libvcs.url import registry as url_tools

from vcspull import exc
from vcspull._internal.private_path import PrivatePath
from vcspull._internal.worktree_sync import (
    WorktreeAction,
    plan_worktree_sync,
    sync_all_worktrees,
)
from vcspull.config import expand_dir, filter_repos, find_config_files, load_configs
from vcspull.types import ConfigDict

from ._colors import Colors, get_color_mode
from ._output import (
    OutputFormatter,
    OutputMode,
    PlanAction,
    PlanEntry,
    PlanRenderOptions,
    PlanResult,
    PlanSummary,
    get_output_mode,
)
from ._workspaces import filter_by_workspace
from .status import check_repo_status

log = logging.getLogger(__name__)

ProgressCallback = Callable[[str, datetime], None]


PLAN_SYMBOLS: dict[PlanAction, str] = {
    PlanAction.CLONE: "+",
    PlanAction.UPDATE: "~",
    PlanAction.UNCHANGED: "✓",
    PlanAction.BLOCKED: "⚠",
    PlanAction.ERROR: "✗",
}

PLAN_ORDER: dict[PlanAction, int] = {
    PlanAction.ERROR: 0,
    PlanAction.BLOCKED: 1,
    PlanAction.CLONE: 2,
    PlanAction.UPDATE: 3,
    PlanAction.UNCHANGED: 4,
}

PLAN_TIP_MESSAGE = (
    "Tip: run without --dry-run to apply. Use --show-unchanged to include ✓ rows."
)

DEFAULT_PLAN_CONCURRENCY = max(1, min(32, (os.cpu_count() or 4) * 2))
ANSI_ESCAPE_RE = re.compile(r"\x1b\[[0-9;]*m")


@dataclass
class SyncPlanConfig:
    """Configuration options for building sync plans."""

    fetch: bool
    offline: bool


def _visible_length(text: str) -> int:
    """Return the printable length of string stripped of ANSI codes."""
    return len(ANSI_ESCAPE_RE.sub("", text))


class PlanProgressPrinter:
    """Render incremental plan progress for human-readable dry runs."""

    def __init__(self, total: int, colors: Colors, enabled: bool) -> None:
        self.total = total
        self._colors = colors
        self._enabled = enabled and total > 0
        self._stream = sys.stdout
        self._last_render_len = 0

    def update(self, summary: PlanSummary, processed: int) -> None:
        """Update the progress line with the latest summary counts."""
        if not self._enabled:
            return

        line = " ".join(
            (
                f"Progress: {processed}/{self.total}",
                self._colors.success(f"+:{summary.clone}"),
                self._colors.warning(f"~:{summary.update}"),
                self._colors.muted(f"✓:{summary.unchanged}"),
                self._colors.warning(f"⚠:{summary.blocked}"),
                self._colors.error(f"✗:{summary.errors}"),
            ),
        )
        clean_len = _visible_length(line)
        padding = max(self._last_render_len - clean_len, 0)
        self._stream.write("\r" + line + " " * padding)
        self._stream.flush()
        self._last_render_len = clean_len

    def finish(self) -> None:
        """Ensure the progress line is terminated with a newline."""
        if not self._enabled:
            return
        self._stream.write("\n")
        self._stream.flush()


def _extract_repo_url(repo: ConfigDict) -> str | None:
    """Extract the primary repository URL from a config dictionary."""
    url = repo.get("url")
    if isinstance(url, str):
        return url
    pip_url = repo.get("pip_url")
    if isinstance(pip_url, str):
        return pip_url
    return None


def _get_repo_path(repo: ConfigDict) -> pathlib.Path:
    """Return the resolved filesystem path for a repository entry."""
    raw_path = repo.get("path")
    if raw_path is None:
        return pathlib.Path.cwd()
    return pathlib.Path(str(raw_path)).expanduser()


def clamp(n: int, _min: int, _max: int) -> int:
    """Clamp a number between a min and max value."""
    return max(_min, min(n, _max))


EXIT_ON_ERROR_MSG = "Exiting via error (--exit-on-error passed)"
NO_REPOS_FOR_TERM_MSG = 'No repo found in config(s) for "{name}"'


def _maybe_fetch(
    repo_path: pathlib.Path,
    *,
    config: SyncPlanConfig,
) -> tuple[bool, str | None]:
    """Optionally fetch remote refs to provide accurate status."""
    if config.offline or not config.fetch:
        return True, None
    if not (repo_path / ".git").exists():
        return True, None

    try:
        result = subprocess.run(
            ["git", "fetch", "--prune"],
            cwd=repo_path,
            capture_output=True,
            text=True,
            check=False,
        )
    except FileNotFoundError:
        return False, "git executable not found"
    except OSError as exc:
        return False, str(exc)

    if result.returncode != 0:
        message = result.stderr.strip() or result.stdout.strip()
        if not message:
            message = f"git fetch failed with exit code {result.returncode}"
        return False, message

    return True, None


def _determine_plan_action(
    status: dict[str, t.Any],
    *,
    config: SyncPlanConfig,
) -> tuple[PlanAction, str | None]:
    """Decide which plan action applies to a repository."""
    if not status.get("exists"):
        return PlanAction.CLONE, "missing"

    if not status.get("is_git"):
        return PlanAction.UPDATE, "non-git VCS (detailed plan not available)"

    clean_state = status.get("clean")
    if clean_state is False:
        return PlanAction.BLOCKED, "working tree has local changes"

    ahead = status.get("ahead")
    behind = status.get("behind")

    if isinstance(ahead, int) and isinstance(behind, int):
        if ahead > 0 and behind > 0:
            return PlanAction.BLOCKED, f"diverged (ahead {ahead}, behind {behind})"
        if behind > 0:
            return PlanAction.UPDATE, f"behind {behind}"
        if ahead > 0:
            return PlanAction.BLOCKED, f"ahead by {ahead}"
        return PlanAction.UNCHANGED, "up to date"

    if config.offline:
        return PlanAction.UPDATE, "remote state unknown (offline)"

    return PlanAction.UPDATE, "remote state unknown; use --fetch"


def _update_summary(summary: PlanSummary, action: PlanAction) -> None:
    """Update summary counters for the given plan action."""
    if action is PlanAction.CLONE:
        summary.clone += 1
    elif action is PlanAction.UPDATE:
        summary.update += 1
    elif action is PlanAction.UNCHANGED:
        summary.unchanged += 1
    elif action is PlanAction.BLOCKED:
        summary.blocked += 1
    elif action is PlanAction.ERROR:
        summary.errors += 1


def _build_plan_entry(
    repo: ConfigDict,
    *,
    config: SyncPlanConfig,
) -> PlanEntry:
    """Construct a plan entry for a repository configuration."""
    repo_path = _get_repo_path(repo)
    workspace_root = str(repo.get("workspace_root", ""))

    fetch_ok = True
    fetch_error: str | None = None
    if repo_path.exists() and (repo_path / ".git").exists():
        fetch_ok, fetch_error = _maybe_fetch(repo_path, config=config)

    status = check_repo_status(repo, detailed=True)

    action: PlanAction
    detail: str | None
    if not fetch_ok:
        action = PlanAction.ERROR
        detail = fetch_error or "failed to refresh remotes"
    else:
        action, detail = _determine_plan_action(status, config=config)

    return PlanEntry(
        name=str(repo.get("name", "unknown")),
        path=str(PrivatePath(repo_path)),
        workspace_root=workspace_root,
        action=action,
        detail=detail,
        url=_extract_repo_url(repo),
        branch=status.get("branch"),
        remote_branch=None,
        current_rev=None,
        target_rev=None,
        ahead=status.get("ahead"),
        behind=status.get("behind"),
        dirty=status.get("clean") is False if status.get("clean") is not None else None,
        error=fetch_error if not fetch_ok else None,
    )


async def _build_plan_result_async(
    repos: list[ConfigDict],
    *,
    config: SyncPlanConfig,
    progress: PlanProgressPrinter | None,
) -> PlanResult:
    """Build a plan asynchronously while updating progress output."""
    if not repos:
        return PlanResult(entries=[], summary=PlanSummary())

    semaphore = asyncio.Semaphore(min(DEFAULT_PLAN_CONCURRENCY, len(repos)))
    entries: list[PlanEntry] = []
    summary = PlanSummary()

    async def evaluate(repo: ConfigDict) -> PlanEntry:
        async with semaphore:
            return await asyncio.to_thread(_build_plan_entry, repo=repo, config=config)

    tasks = [asyncio.create_task(evaluate(repo)) for repo in repos]

    for index, task in enumerate(asyncio.as_completed(tasks), start=1):
        entry = await task
        entries.append(entry)
        _update_summary(summary, entry.action)
        if progress is not None:
            progress.update(summary, index)

    return PlanResult(entries=entries, summary=summary)


def _filter_entries_for_display(
    entries: list[PlanEntry],
    *,
    show_unchanged: bool,
) -> list[PlanEntry]:
    """Filter entries based on whether unchanged repos should be rendered."""
    if show_unchanged:
        return list(entries)
    return [entry for entry in entries if entry.action is not PlanAction.UNCHANGED]


def _format_detail_text(
    entry: PlanEntry,
    *,
    colors: Colors,
    include_extras: bool,
) -> str:
    """Generate the detail text for a plan entry."""
    detail = entry.detail or ""
    extra_bits: list[str] = []

    if include_extras:
        if entry.action is PlanAction.UPDATE and entry.behind:
            extra_bits.append(f"behind {entry.behind}")
        if entry.action is PlanAction.CLONE and entry.url:
            extra_bits.append(entry.url)
        if entry.action is PlanAction.BLOCKED and entry.error:
            extra_bits.append(entry.error)

    if extra_bits:
        detail = f"{detail} {'; '.join(extra_bits)}".strip()

    color_map: dict[PlanAction, t.Callable[[str], str]] = {
        PlanAction.CLONE: colors.success,
        PlanAction.UPDATE: colors.warning,
        PlanAction.UNCHANGED: colors.muted,
        PlanAction.BLOCKED: colors.warning,
        PlanAction.ERROR: colors.error,
    }

    formatter = color_map.get(entry.action, colors.info)
    return formatter(detail) if detail else ""


def _render_plan(
    formatter: OutputFormatter,
    colors: Colors,
    plan: PlanResult,
    render_options: PlanRenderOptions,
    *,
    dry_run: bool,
    total_repos: int,
) -> None:
    """Render the plan in human-readable format."""
    summary = plan.summary
    summary_line = (
        f"Plan: "
        f"{colors.success(str(summary.clone))} to clone (+), "
        f"{colors.warning(str(summary.update))} to update (~), "
        f"{colors.muted(str(summary.unchanged))} unchanged (✓), "
        f"{colors.warning(str(summary.blocked))} blocked (⚠), "
        f"{colors.error(str(summary.errors))} errors (✗)"
    )
    formatter.emit_text(summary_line)

    if total_repos == 0:
        formatter.emit_text(colors.warning("No repositories matched the criteria."))
        return

    if render_options.summary_only:
        if dry_run:
            formatter.emit_text(colors.muted(PLAN_TIP_MESSAGE))
        return

    display_entries = _filter_entries_for_display(
        sorted(
            plan.entries,
            key=lambda entry: (
                PLAN_ORDER.get(entry.action, 99),
                entry.workspace_root or "",
                entry.name.lower(),
            ),
        ),
        show_unchanged=render_options.show_unchanged,
    )

    if not display_entries:
        formatter.emit_text(colors.muted("All repositories are up to date."))
        if dry_run:
            formatter.emit_text(colors.muted(PLAN_TIP_MESSAGE))
        return

    formatter.emit_text("")

    grouped: dict[str, list[PlanEntry]] = {}
    for entry in display_entries:
        key = entry.workspace_root or "(no workspace)"
        grouped.setdefault(key, []).append(entry)

    for idx, (workspace, group_entries) in enumerate(grouped.items()):
        if idx > 0:
            formatter.emit_text("")
        formatter.emit_text(colors.highlight(workspace))
        name_width = max(len(entry.name) for entry in group_entries)

        for entry in group_entries:
            symbol = PLAN_SYMBOLS.get(entry.action, "?")
            color_map: dict[PlanAction, t.Callable[[str], str]] = {
                PlanAction.CLONE: colors.success,
                PlanAction.UPDATE: colors.warning,
                PlanAction.UNCHANGED: colors.muted,
                PlanAction.BLOCKED: colors.warning,
                PlanAction.ERROR: colors.error,
            }
            symbol_text = color_map.get(entry.action, colors.info)(symbol)

            display_path = entry.path
            if render_options.relative_paths and entry.workspace_root:
                workspace_path = pathlib.Path(entry.workspace_root).expanduser()
                try:
                    rel_path = pathlib.Path(entry.path).relative_to(workspace_path)
                    display_path = str(rel_path)
                except ValueError:
                    display_path = entry.path
            else:
                # Contract home directory for privacy/brevity in human output
                display_path = str(PrivatePath(display_path))

            detail_text = _format_detail_text(
                entry,
                colors=colors,
                include_extras=render_options.verbosity > 0 or render_options.long,
            )

            line = (
                f"  {symbol_text} {colors.info(entry.name.ljust(name_width))}  "
                f"{colors.muted(display_path)}"
            )
            if detail_text:
                line = f"{line}  {detail_text}"
            formatter.emit_text(line.rstrip())

            if render_options.long or render_options.verbosity > 1:
                extra_lines: list[str] = []
                if entry.url:
                    extra_lines.append(f"url: {entry.url}")
                if entry.ahead is not None or entry.behind is not None:
                    extra_lines.append(
                        f"ahead/behind: {entry.ahead or 0}/{entry.behind or 0}",
                    )
                if entry.error:
                    extra_lines.append(f"error: {entry.error}")
                for msg in extra_lines:
                    formatter.emit_text(f"    {colors.muted(msg)}")

    if dry_run:
        formatter.emit_text(colors.muted(PLAN_TIP_MESSAGE))


def _emit_plan_output(
    formatter: OutputFormatter,
    colors: Colors,
    plan: PlanResult,
    render_options: PlanRenderOptions,
    *,
    dry_run: bool,
    total_repos: int,
) -> None:
    """Emit plan output for the requested format."""
    if formatter.mode == OutputMode.HUMAN:
        _render_plan(
            formatter=formatter,
            colors=colors,
            plan=plan,
            render_options=render_options,
            dry_run=dry_run,
            total_repos=total_repos,
        )
        return

    display_entries = _filter_entries_for_display(
        plan.entries,
        show_unchanged=render_options.show_unchanged,
    )

    for entry in display_entries:
        formatter.emit(entry)
    formatter.emit(plan.summary)


def create_sync_subparser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Create ``vcspull sync`` argument subparser."""
    config_file = parser.add_argument(
        "-f",
        "--file",
        dest="config",
        metavar="FILE",
        help="path to config file (default: ~/.vcspull.yaml or ./.vcspull.yaml)",
    )
    parser.add_argument(
        "-w",
        "--workspace",
        "--workspace-root",
        dest="workspace_root",
        metavar="DIR",
        help="filter by workspace root directory",
    )
    parser.add_argument(
        "repo_patterns",
        metavar="pattern",
        nargs="*",
        help="patterns / terms of repos, accepts globs / fnmatch(3)",
    )
    parser.add_argument(
        "--dry-run",
        "-n",
        action="store_true",
        help="preview what would be synced without making changes",
    )
    parser.add_argument(
        "--json",
        action="store_true",
        dest="output_json",
        help="output as JSON",
    )
    parser.add_argument(
        "--ndjson",
        action="store_true",
        dest="output_ndjson",
        help="output as NDJSON (one JSON per line)",
    )
    parser.add_argument(
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="when to use colors (default: auto)",
    )
    parser.add_argument(
        "--exit-on-error",
        "-x",
        action="store_true",
        dest="exit_on_error",
        help="exit immediately encountering error (when syncing multiple repos)",
    )
    parser.add_argument(
        "--show-unchanged",
        action="store_true",
        help="include repositories that are already up to date",
    )
    parser.add_argument(
        "--summary-only",
        action="store_true",
        dest="summary_only",
        help="print only the plan summary line",
    )
    parser.add_argument(
        "--long",
        action="store_true",
        dest="long_view",
        help="show extended details for each repository",
    )
    parser.add_argument(
        "--relative-paths",
        action="store_true",
        dest="relative_paths",
        help="display repository paths relative to the workspace root",
    )
    parser.add_argument(
        "--fetch",
        action="store_true",
        help="refresh remote tracking information before planning",
    )
    parser.add_argument(
        "--offline",
        action="store_true",
        help="skip network access while planning (overrides --fetch)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbosity",
        default=0,
        help="increase plan verbosity (-vv for maximum detail)",
    )
    parser.add_argument(
        "--all",
        "-a",
        action="store_true",
        dest="sync_all",
        help="sync all configured repositories",
    )
    parser.add_argument(
        "--include-worktrees",
        action="store_true",
        dest="include_worktrees",
        help="also sync configured worktrees for each repository",
    )

    try:
        import shtab

        config_file.complete = shtab.FILE  # type: ignore
    except ImportError:
        pass
    return parser


def sync(
    repo_patterns: list[str],
    config: pathlib.Path | None,
    workspace_root: str | None,
    dry_run: bool,
    output_json: bool,
    output_ndjson: bool,
    color: str,
    exit_on_error: bool,
    show_unchanged: bool,
    summary_only: bool,
    long_view: bool,
    relative_paths: bool,
    fetch: bool,
    offline: bool,
    verbosity: int,
    sync_all: bool = False,
    parser: argparse.ArgumentParser
    | None = None,  # optional so sync can be unit tested
    include_worktrees: bool = False,
) -> None:
    """Entry point for ``vcspull sync``."""
    # Show help if no patterns and --all not specified
    if not repo_patterns and not sync_all:
        if parser is not None:
            parser.print_help()
        else:
            log.warning(
                "sync() called with no repo patterns and no --all flag; nothing to do",
            )
        return

    output_mode = get_output_mode(output_json, output_ndjson)
    formatter = OutputFormatter(output_mode)
    colors = Colors(get_color_mode(color))

    verbosity_level = clamp(verbosity, 0, 2)
    render_options = PlanRenderOptions(
        show_unchanged=show_unchanged,
        summary_only=summary_only,
        long=long_view,
        verbosity=verbosity_level,
        relative_paths=relative_paths,
    )
    plan_config = SyncPlanConfig(fetch=bool(fetch and not offline), offline=offline)

    if config:
        configs = load_configs([config])
    else:
        configs = load_configs(find_config_files(include_home=True))
    found_repos: list[ConfigDict] = []
    unmatched_count = 0

    if sync_all:
        if repo_patterns:
            msg = "--all cannot be combined with positional patterns"
            if parser is not None:
                parser.error(msg)
            else:
                raise SystemExit(msg)
        # Load all repos when --all is specified
        found_repos = list(configs)
    else:
        for repo_pattern in repo_patterns:
            path, vcs_url, name = None, None, None
            if any(repo_pattern.startswith(n) for n in ["./", "/", "~", "$HOME"]):
                path = repo_pattern
            elif any(repo_pattern.startswith(n) for n in ["http", "git", "svn", "hg"]):
                vcs_url = repo_pattern
            else:
                name = repo_pattern

            found = filter_repos(configs, path=path, vcs_url=vcs_url, name=name)
            if not found:
                search_term = name or path or vcs_url or repo_pattern
                log.debug(NO_REPOS_FOR_TERM_MSG.format(name=search_term))
                if not summary_only:
                    formatter.emit_text(
                        f"{colors.error('✗')} "
                        f"{NO_REPOS_FOR_TERM_MSG.format(name=search_term)}",
                    )
                unmatched_count += 1
            found_repos.extend(found)

        # Deduplicate repos matched by multiple patterns
        seen_paths: set[str] = set()
        deduped: list[ConfigDict] = []
        for repo in found_repos:
            key = str(repo.get("path", ""))
            if key not in seen_paths:
                seen_paths.add(key)
                deduped.append(repo)
        found_repos = deduped

    if workspace_root:
        found_repos = filter_by_workspace(found_repos, workspace_root)

    total_repos = len(found_repos)

    if dry_run:
        progress_enabled = formatter.mode == OutputMode.HUMAN and sys.stdout.isatty()
        progress_printer = PlanProgressPrinter(total_repos, colors, progress_enabled)
        start_time = perf_counter()
        plan_result = asyncio.run(
            _build_plan_result_async(
                found_repos,
                config=plan_config,
                progress=progress_printer if progress_enabled else None,
            ),
        )
        plan_result.summary.duration_ms = int((perf_counter() - start_time) * 1000)
        if progress_enabled:
            progress_printer.finish()
        _emit_plan_output(
            formatter=formatter,
            colors=colors,
            plan=plan_result,
            render_options=render_options,
            dry_run=True,
            total_repos=total_repos,
        )

        # Show worktree plans if --include-worktrees is set
        if include_worktrees:
            for repo in found_repos:
                worktrees_config = repo.get("worktrees")
                if not worktrees_config:
                    continue

                repo_name = str(repo.get("name", "unknown"))
                repo_path = _get_repo_path(repo)
                workspace_label = str(repo.get("workspace_root", ""))
                workspace_path = expand_dir(pathlib.Path(workspace_label))

                wt_entries = plan_worktree_sync(
                    repo_path, worktrees_config, workspace_path
                )

                for entry in wt_entries:
                    ref_display = f"{entry.ref_type}:{entry.ref_value}"
                    wt_path_display = str(PrivatePath(entry.worktree_path))

                    action_symbols = {
                        WorktreeAction.CREATE: colors.success("+"),
                        WorktreeAction.UPDATE: colors.warning("~"),
                        WorktreeAction.UNCHANGED: colors.muted("✓"),
                        WorktreeAction.BLOCKED: colors.warning("⚠"),
                        WorktreeAction.ERROR: colors.error("✗"),
                    }
                    sym = action_symbols.get(entry.action, "?")
                    ref = colors.info(ref_display)
                    detail = entry.detail or entry.error or ""

                    formatter.emit_text(
                        f"  {sym} worktree {colors.info(repo_name)} "
                        f"{ref} {colors.muted('→')} {wt_path_display}"
                        f"  {colors.muted(detail)}".rstrip(),
                    )

        formatter.finalize()
        return

    if total_repos == 0:
        if unmatched_count > 0:
            summary = {
                "total": 0,
                "synced": 0,
                "previewed": 0,
                "failed": 0,
                "unmatched": unmatched_count,
            }
            _emit_summary(formatter, colors, summary)
            if exit_on_error:
                formatter.finalize()
                if parser is not None:
                    parser.exit(status=1, message=EXIT_ON_ERROR_MSG)
                raise SystemExit(EXIT_ON_ERROR_MSG)
        else:
            formatter.emit_text(
                colors.warning("No repositories matched the criteria."),
            )
        formatter.finalize()
        return

    is_human = formatter.mode == OutputMode.HUMAN

    summary = {
        "total": 0,
        "synced": 0,
        "previewed": 0,
        "failed": 0,
        "unmatched": unmatched_count,
    }

    progress_callback: ProgressCallback
    if is_human:
        progress_callback = progress_cb
    else:

        def silent_progress(output: str, timestamp: datetime) -> None:
            """Suppress progress for machine-readable output."""
            return

        progress_callback = silent_progress

    for repo in found_repos:
        repo_name = repo.get("name", "unknown")
        repo_path = repo.get("path", "unknown")
        workspace_label = repo.get("workspace_root", "")
        display_repo_path = str(PrivatePath(repo_path))

        summary["total"] += 1

        event: dict[str, t.Any] = {
            "reason": "sync",
            "name": repo_name,
            "path": display_repo_path,
            "workspace_root": str(workspace_label),
        }

        buffer: StringIO | None = None
        captured_output: str | None = None
        try:
            if is_human:
                update_repo(repo, progress_callback=progress_callback)
            else:
                buffer = StringIO()
                with (
                    contextlib.redirect_stdout(buffer),
                    contextlib.redirect_stderr(
                        buffer,
                    ),
                ):
                    update_repo(repo, progress_callback=progress_callback)
                captured_output = buffer.getvalue()
        except Exception as e:
            summary["failed"] += 1
            event["status"] = "error"
            event["error"] = str(e)
            if not is_human and buffer is not None and not captured_output:
                captured_output = buffer.getvalue()
            if captured_output:
                event["details"] = captured_output.strip()
            formatter.emit(event)
            if is_human:
                log.debug(
                    "Failed syncing %s",
                    repo_name,
                )
            if log.isEnabledFor(logging.DEBUG):
                import traceback

                traceback.print_exc()
            formatter.emit_text(
                f"{colors.error('✗')} Failed syncing {colors.info(repo_name)}: "
                f"{colors.error(str(e))}",
            )
            if exit_on_error:
                _emit_summary(formatter, colors, summary)
                formatter.finalize()
                if parser is not None:
                    parser.exit(status=1, message=EXIT_ON_ERROR_MSG)
                raise SystemExit(EXIT_ON_ERROR_MSG) from e
            continue

        summary["synced"] += 1
        event["status"] = "synced"
        formatter.emit(event)
        formatter.emit_text(
            f"{colors.success('✓')} Synced {colors.info(repo_name)} "
            f"{colors.muted('→')} {display_repo_path}",
        )

        # Sync worktrees if enabled and configured
        worktrees_config = repo.get("worktrees")
        if include_worktrees and worktrees_config:
            workspace_path = expand_dir(pathlib.Path(str(workspace_label)))
            repo_path_obj = pathlib.Path(str(repo_path))

            wt_result = sync_all_worktrees(
                repo_path_obj,
                worktrees_config,
                workspace_path,
                dry_run=dry_run,
            )

            for entry in wt_result.entries:
                ref_display = f"{entry.ref_type}:{entry.ref_value}"
                wt_path_display = str(PrivatePath(entry.worktree_path))

                if entry.action == WorktreeAction.CREATE:
                    sym = colors.success("+")
                    ref = colors.info(ref_display)
                    arrow = colors.muted("→")
                    formatter.emit_text(
                        f"    {sym} worktree {ref} {arrow} {wt_path_display}",
                    )
                elif entry.action == WorktreeAction.UPDATE:
                    sym = colors.warning("~")
                    ref = colors.info(ref_display)
                    arrow = colors.muted("→")
                    formatter.emit_text(
                        f"    {sym} worktree {ref} {arrow} {wt_path_display}",
                    )
                elif entry.action == WorktreeAction.BLOCKED:
                    sym = colors.warning("⚠")
                    ref = colors.info(ref_display)
                    formatter.emit_text(
                        f"    {sym} worktree {ref} blocked: {entry.detail}",
                    )
                elif entry.action == WorktreeAction.ERROR:
                    formatter.emit_text(
                        f"    {colors.error('✗')} worktree {colors.info(ref_display)} "
                        f"error: {entry.error}",
                    )

            # Tally worktree results into summary
            summary["worktree_created"] = (
                summary.get("worktree_created", 0) + wt_result.created
            )
            summary["worktree_updated"] = (
                summary.get("worktree_updated", 0) + wt_result.updated
            )
            summary["worktree_failed"] = (
                summary.get("worktree_failed", 0) + wt_result.errors
            )
            # Count worktree errors as failures for exit code
            summary["failed"] += wt_result.errors

            if exit_on_error and wt_result.errors > 0:
                _emit_summary(formatter, colors, summary)
                formatter.finalize()
                if parser is not None:
                    parser.exit(status=1, message=EXIT_ON_ERROR_MSG)
                raise SystemExit(EXIT_ON_ERROR_MSG)

    _emit_summary(formatter, colors, summary)

    if exit_on_error and unmatched_count > 0:
        formatter.finalize()
        if parser is not None:
            parser.exit(status=1, message=EXIT_ON_ERROR_MSG)
        raise SystemExit(EXIT_ON_ERROR_MSG)

    formatter.finalize()


def _emit_summary(
    formatter: OutputFormatter,
    colors: Colors,
    summary: dict[str, int],
) -> None:
    """Emit the structured summary event and optional human-readable text."""
    formatter.emit({"reason": "summary", **summary})
    if formatter.mode == OutputMode.HUMAN:
        previewed = summary.get("previewed", 0)
        unmatched = summary.get("unmatched", 0)
        parts = [
            f"\n{colors.info('Summary:')} "
            f"{colors.info(str(summary['total']))} repos, "
            f"{colors.success(str(summary['synced']))} synced, "
            f"{colors.error(str(summary['failed']))} failed",
        ]
        if previewed > 0:
            parts.append(
                f", {colors.warning(str(previewed))} previewed",
            )
        if unmatched > 0:
            parts.append(
                f", {colors.warning(str(unmatched))} unmatched",
            )
        wt_created = summary.get("worktree_created", 0)
        wt_updated = summary.get("worktree_updated", 0)
        wt_failed = summary.get("worktree_failed", 0)
        if wt_created or wt_updated or wt_failed:
            wt_parts = []
            if wt_created:
                wt_parts.append(f"{colors.success(str(wt_created))} created")
            if wt_updated:
                wt_parts.append(f"{colors.warning(str(wt_updated))} updated")
            if wt_failed:
                wt_parts.append(f"{colors.error(str(wt_failed))} failed")
            parts.append(
                f"; worktrees: {', '.join(wt_parts)}",
            )
        formatter.emit_text("".join(parts))


def progress_cb(output: str, timestamp: datetime) -> None:
    """CLI Progress callback for command."""
    sys.stdout.write(output)
    sys.stdout.flush()


def guess_vcs(url: str) -> VCSLiteral | None:
    """Guess the VCS from a URL."""
    vcs_matches = url_tools.registry.match(url=url, is_explicit=True)

    if len(vcs_matches) == 0:
        log.warning("No vcs found for %s", url)
        return None
    if len(vcs_matches) > 1:
        log.warning("No exact matches for %s", url)
        return None

    return t.cast("VCSLiteral", vcs_matches[0].vcs)


class CouldNotGuessVCSFromURL(exc.VCSPullException):
    """Raised when no VCS could be guessed from a URL."""

    def __init__(self, repo_url: str, *args: object, **kwargs: object) -> None:
        return super().__init__(f"Could not automatically determine VCS for {repo_url}")


class SyncFailedError(exc.VCSPullException):
    """Raised when a sync operation completes but with errors."""

    def __init__(
        self,
        repo_name: str,
        errors: str,
        *args: object,
        **kwargs: object,
    ) -> None:
        self.repo_name = repo_name
        self.errors = errors
        message = f"Sync failed for {repo_name}"
        if errors:
            message = f"{message}: {errors}"
        super().__init__(message)


def update_repo(
    repo_dict: t.Any,
    progress_callback: ProgressCallback | None = None,
    # repo_dict: Dict[str, Union[str, Dict[str, GitRemote], pathlib.Path]]
) -> GitSync | HgSync | SvnSync:
    """Synchronize a single repository."""
    repo_dict = deepcopy(repo_dict)
    if "pip_url" not in repo_dict:
        repo_dict["pip_url"] = repo_dict.pop("url")
    if "url" not in repo_dict:
        repo_dict["url"] = repo_dict.pop("pip_url")

    repo_dict["progress_callback"] = progress_callback or progress_cb

    if repo_dict.get("vcs") is None:
        vcs = guess_vcs(url=repo_dict["url"])
        if vcs is None:
            raise CouldNotGuessVCSFromURL(repo_url=repo_dict["url"])

        repo_dict["vcs"] = vcs

    r: GitSync | HgSync | SvnSync = create_project(**repo_dict)
    if repo_dict.get("vcs") == "git":
        result = r.update_repo(set_remotes=True)
    else:
        result = r.update_repo()

    if result is not None and not result.ok:
        error_messages = "; ".join(e.message for e in result.errors)
        repo_name = str(repo_dict.get("name", repo_dict.get("url", "unknown")))
        raise SyncFailedError(repo_name=repo_name, errors=error_messages)

    return r



#!/usr/bin/env python
"""Manage multiple git, mercurial, svn repositories from a YAML / JSON file.

:copyright: Copyright 2013-2018 Tony Narlock.
:license: MIT, see LICENSE for details
"""

# Set default logging handler to avoid "No handler found" warnings.
from __future__ import annotations

import logging
from logging import NullHandler

from . import cli

logging.getLogger(__name__).addHandler(NullHandler())
